#!/usr/bin/perl

minmeanmax("nc93");
minmeanmax("nc94");

large("nc93", "Host");
large("nc94", "Emulator");

small("nc93", "Host");
small("nc94", "Emulator");


sub minmeanmax {
    my ($machine) = @_;

    my $infile = "data.$machine";

    my $xrange = 345600;
    my $min;
    my $max;
    my $sum = 0;
    my $num = 0;

    open(IN, "$infile") || die "$infile";

  LINE: while (<IN>) {
      my ($serial, $time_sec, $time_cyc, $cps) = split;

      last LINE if ($time_sec > $xrange);

      $sum += $cps;
      $num++;
      if (!defined($min) || $cps < $min) {
	  $min = $cps;
      }
      if (!defined($max) || $cps > $max) {
	  $max = $cps;
      }
  }

    close(IN);

    my $avg = $sum / $num;
    my $minpct = ($min - $avg) / $avg * 100;
    my $maxpct = ($max - $avg) / $avg * 100;

    print "$machine: sum $sum\n";
    print "$machine: num $num\n";
    print "$machine: avg $avg\n";
    print "$machine: min $min ($minpct %)\n";
    print "$machine: max $max ($maxpct %)\n";
}



sub large {
    my ($machine, $name) = @_;

    my $infile = "data.$machine";
    my $outfile = "$infile.large";
    my $epsfile = "$outfile.eps";

    my $xrange = 345600;

    open(IN, "$infile") || die "$infile";
    open(OUT, ">$outfile") || die "$outfile";

    my $count = 0;

  LINE: while (<IN>) {
      my ($serial, $time_sec, $time_cyc, $cps) = split;

      last LINE if ($time_sec > $xrange);

      if ($count-- <= 0 || $serial < 100) {
	  print OUT;
	  $count = 100;
      }
  }

    close(OUT);
    close(IN);

    open(GNU, "|gnuplot > $epsfile") || die "$epsfile";

    print GNU "set term postscript landscape color \"Helvetica\" 18 \n";
    print GNU "set title '$name system calibration' \n";
    print GNU "set xlabel 'Elapsed time (seconds)' \n";
    print GNU "set ylabel 'Measured CPS' \n";
    print GNU "set format y \"%12.f\" \n";
    print GNU "set xtics 0,86400,$xrange \n";
    print GNU "set xrange [-5000:$xrange] \n";
    print GNU "plot '$outfile' using 2:4 title '' with lines lt -1 lw 1 \n";

    close(GNU);

    system"gv $epsfile &";
}



sub small {
    my ($machine, $name) = @_;

    my $infile = "data.$machine";
    my $outfile = "$infile.small";
    my $epsfile = "$outfile.eps";

    my $xrange = 2500;

    open(IN, "$infile") || die "$infile";
    open(OUT, ">$outfile") || die "$outfile";

    my $count = 0;

  LINE: while (<IN>) {
      my ($serial, $time_sec, $time_cyc, $cps) = split;

      last LINE if ($time_sec > $xrange);

      print OUT;
  }

    close(OUT);
    close(IN);

    open(GNU, "|gnuplot > $epsfile") || die "$epsfile";

    print GNU "set term postscript landscape color \"Helvetica\" 18 \n";
    print GNU "set title '$name system calibration' \n";
    print GNU "set xlabel 'Elapsed time (seconds)' \n";
    print GNU "set ylabel 'Measured CPS' \n";
    print GNU "set format y \"%12.f\" \n";
    print GNU "set xrange [-50:$xrange] \n";
    print GNU "plot '$outfile' using 2:4 title '' with lines lt -1 lw 1 \n";

    close(GNU);

    system"gv $epsfile &";
}
