/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

static inline void
time_stamp (u_int64_t *d)
{
  asm volatile ("rdtsc" : "=&A" (*d));
}


int
main (int argc, char **argv)
{
  u_int64_t start_tsc;
  struct timeval start_tv;
  long double start_time;
  u_int64_t now_tsc;
  struct timeval now_tv;
  long double now_time;
  long double diff;
  int i;

  /* Warm-up phase */

  for (i=0; i<10000; i++) {
    time_stamp(&start_tsc);
    gettimeofday(&start_tv, NULL);
    start_time = start_tv.tv_sec + start_tv.tv_usec / 1000000.0;
  }

  /* Take the initial reference time */

  sleep(1);
  time_stamp(&start_tsc);
  gettimeofday(&start_tv, NULL);
  start_time = start_tv.tv_sec + start_tv.tv_usec / 1000000.0;

  /* Take subsequent times and output difference */

  for (i = 1; /* infinite loop */; i++) {
    sleep(1);
    time_stamp(&now_tsc);
    gettimeofday(&now_tv, NULL);
    now_time = now_tv.tv_sec + now_tv.tv_usec / 1000000.0;

    diff = now_time - start_time;

    /* Format: A B C D\n
     *  A: serial number [int]
     *  B: elapsed time (in seconds) [float]
     *  C: elapsed cycle count [int]
     *  D: measured cycles per second [float]
     */

    printf("%7d %14.6Lf %lld %Lf\n", 
	   i, 
	   diff, 
	   (now_tsc - start_tsc), 
	   (now_tsc - start_tsc) / diff);
    fflush(stdout);
  }

  return 0;
}


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

static inline void
time_stamp (u_int64_t *d)
{
  asm volatile ("rdtsc" : "=&A" (*d));
}


int
main (int argc, char **argv)
{
  u_int64_t start_tsc;
  struct timeval start_tv;
  long double start_time;
  u_int64_t now_tsc;
  struct timeval now_tv;
  long double now_time;
  long double diff;
  int i;

  /* Warm-up phase */

  for (i=0; i<10000; i++) {
    time_stamp(&start_tsc);
    gettimeofday(&start_tv, NULL);
    start_time = start_tv.tv_sec + start_tv.tv_usec / 1000000.0;
  }

  /* Take the initial reference time */

  sleep(1);
  time_stamp(&start_tsc);
  gettimeofday(&start_tv, NULL);
  start_time = start_tv.tv_sec + start_tv.tv_usec / 1000000.0;

  /* Take subsequent times and output difference */

  for (i = 1; /* infinite loop */; i++) {
    sleep(1);
    time_stamp(&now_tsc);
    gettimeofday(&now_tv, NULL);
    now_time = now_tv.tv_sec + now_tv.tv_usec / 1000000.0;

    diff = now_time - start_time;

    /* Format: A B C D\n
     *  A: serial number [int]
     *  B: elapsed time (in seconds) [float]
     *  C: elapsed cycle count [int]
     *  D: measured cycles per second [float]
     */

    printf("%7d %14.6Lf %lld %Lf\n", 
	   i, 
	   diff, 
	   (now_tsc - start_tsc), 
	   (now_tsc - start_tsc) / diff);
    fflush(stdout);
  }

  return 0;
}

