#!/usr/bin/perl

use English;
use FileHandle;
require "subroutines.pm";

$script_dir = "final-compensate";


#$run_lookahead = 3;
$run_lookahead_adaptec = 1;
$run_lookahead_qlogic = 1;

#$run_lookfar = 3;
$run_lookfar_adaptec = 1;
$run_lookfar_qlogic = 1;

#$run_noskew_nolook = 3;
#$run_noskew_nolook_adaptec = 1;
$run_noskew_nolook_qlogic = 1;

$run_noskew_look = 3;
#$run_noskew_look_adaptec = 1;
$run_noskew_look_qlogic = 1;

#$run_skew_look = 3;
$run_skew_look_adaptec = 1;
$run_skew_look_qlogic = 1;


#$run_fiftyKrpm_lookahead = 3;
$run_fiftyKrpm_lookahead_adaptec = 1;

#$run_fiftyKrpm_noskew_look = 1;
$run_fiftyKrpm_noskew_look_adaptec = 1;


#$run_mems_lookahead = 3;
$run_mems_lookahead_adaptec = 1;

#$run_mems_noskew_look = 3;
$run_mems_noskew_look_adaptec = 1;



$lookahead_us    = 0;
$skew_adaptec_us = 0;
$skew_qlogic_us  = 0;


########################################################################

sub prepare {
    $runid = `date +%s`; chomp ($runid);
    $rundir = "runs/$runstr/$runid";
    system "mkdir -p $rundir";

    @stamps = ();
    $stampsfile = "$rundir/stamps";
    $sourcefile = "$rundir/source";

    open (STAMPS, ">$stampsfile") || die "$stampsfile";
    open (SOURCE, ">$sourcefile") || die "$sourcefile";

    autoflush STAMPS 1;
    autoflush SOURCE 1;

    open_log($rundir);

    print SOURCE "% $runstr\n";
    print SOURCE "% Begin: " . `date`;
}


sub adaptec_connect {
    ## connect emulator
    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_add_single_device();
    targ_stop_memulator();
}


sub qlogic_connect {
    ## connect emulator
    host_rmmod_qla2200();
    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_insmod_qla2200();
    targ_stop_memulator();
}


sub config_cheetah {
    $lookahead_us    = 30;
    $skew_adaptec_us = 60;
    $skew_qlogic_us  = 151;
    $command = "ln -sf cheetah.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    print "$command\n";
    system $command;
}


sub config_fiftyKrpm {
    $lookahead_us    = 63;
    $command = "ln -sf fiftyKrpm.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    print "$command\n";
    system $command;
}


sub config_mems {
    $lookahead_us    = 26;
    $command = "ln -sf mems.g2.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    print "$command\n";
    system $command;
}


sub trace_replay {
    $timestamp = `date +%s`; chomp($timestamp);
    mkdir "$results_dir/$timestamp", 0777 
	or die "Cannot mkdir $results_dir/$timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    print STAMPS "$timestamp\n";
    print SOURCE "\\newcommand{$latexcmd}{$timestamp} % $trace $runstr $extraflags\n";

    ## start timepoint collection
    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);

    ssh_cmd_checkresult("$host_name", "(cd $main_dir/$script_dir ; $main_dir/src/disksim/dixtrac/dxreplay -s -m $trace /dev/sg3)", 0);

    ## stop timepoint collection
    host_rmmod_hosttimes();
    targ_stop_memulator();

    sleep 2;
}


sub full_trace_replay {
    if (1) {
	$latexcmd = "\\timingTwoR";
	$trace = "t-2-r";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingTwoW";
	$trace = "t-2-w";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingEigW";
	$trace = "t-8-w";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingSxtR";
	$trace = "t-64-r";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingSxtW";
	$trace = "t-64-w";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingHndR";
	$trace = "t-128-r";
	trace_replay();
    }

    if (1) {
	$latexcmd = "\\timingHndW";
	$trace = "t-128-w";
	trace_replay();
    }
}


sub adaptec_disconnect {
    ## disconnect emulator
    targ_start_memulator_notimes();
    host_remove_single_device();
    targ_stop_memulator();
}


sub qlogic_disconnect {
    ## disconnect emulator
    targ_start_memulator_notimes();
    host_rmmod_qla2200();
    targ_stop_memulator();
}


sub complete {
    close_log();
    close (SOURCE);
    close (STAMPS);
}


## lookahead (0->60, mem-sim)

for ($outerloop = $run_lookahead;
     $outerloop > 0;
     $outerloop--) {

    system "./make-100-traces.pl";

    ## adaptec

    for ($innerloop = $run_lookahead_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "lookahead-a";
	config_cheetah();
	config_adaptec();
	prepare();
	adaptec_connect();

	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 60; $i++) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	adaptec_disconnect();
	complete();
    }


    ## qlogic

    for ($innerloop = $run_lookahead_qlogic;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "lookahead-q";
	config_cheetah();
	config_qlogic();
	prepare();
	qlogic_connect();
    
	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 60; $i++) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	qlogic_disconnect();
	complete();
    }
}


## lookfar (0->400, hos-mem)

for ($outerloop = $run_lookfar;
     $outerloop > 0;
     $outerloop--) {

    system "./make-100-traces.pl";

    ## adaptec

    for ($innerloop = $run_lookfar_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "lookfar-a";
	config_cheetah();
	config_adaptec();
	prepare();
	adaptec_connect();

	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 400; $i += 10) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	adaptec_disconnect();
	complete();
    }


    ## qlogic

    for ($innerloop = $run_lookfar_qlogic;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "lookfar-q";
	config_cheetah();
	config_qlogic();
	prepare();
	qlogic_connect();
    
	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 400; $i += 10) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	qlogic_disconnect();
	complete();
    }
}


## noskew-nolook

for ($outerloop = $run_noskew_nolook;
     $outerloop > 0;
     $outerloop--) {

    system "make-1000-traces.pl";

    ## adaptec

    for ($innerloop = $run_noskew_nolook_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "noskew-nolook-a";
	config_cheetah();
	config_adaptec();
	prepare();
	adaptec_connect();

	$extraflags = ();

	if (1) {
	    $latexcmd = "\\timingTwoR";
	    $trace = "t-2-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingTwoW";
	    $trace = "t-2-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingEigR";
	    $trace = "t-8-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingEigW";
	    $trace = "t-8-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingSxtR";
	    $trace = "t-64-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingSxtW";
	    $trace = "t-64-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingHndR";
	    $trace = "t-128-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingHndW";
	    $trace = "t-128-w";
	    trace_replay();
	}

	adaptec_disconnect();
	complete();
    }

    ## qlogic

    for ($innerloop = $run_noskew_nolook_qlogic;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "noskew-nolook-q";
	config_cheetah();
	config_qlogic();
	prepare();
	qlogic_connect();

	$extraflags = ();

	if (1) {
	    $latexcmd = "\\timingTwoR";
	    $trace = "t-2-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingTwoW";
	    $trace = "t-2-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingEigR";
	    $trace = "t-8-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingEigW";
	    $trace = "t-8-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingSxtR";
	    $trace = "t-64-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingSxtW";
	    $trace = "t-64-w";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingHndR";
	    $trace = "t-128-r";
	    trace_replay();
	}

	if (1) {
	    $latexcmd = "\\timingHndW";
	    $trace = "t-128-w";
	    trace_replay();
	}

	qlogic_disconnect();
	complete();
    }

}


## noskew-look

for ($outerloop = $run_noskew_look;
     $outerloop > 0;
     $outerloop--) {

    system "make-1000-traces.pl";

    ## adaptec

    for ($innerloop = $run_noskew_look_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "noskew-look-a";
	config_cheetah();
	config_adaptec();
	prepare();
	adaptec_connect();

	$extraflags = "-l $lookahead_us";
	full_trace_replay();

	adaptec_disconnect();
	complete();
    }

    ## qlogic

    for ($innerloop = $run_noskew_look_qlogic;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "noskew-look-q";
	config_cheetah();
	config_qlogic();
	prepare();
	qlogic_connect();

	$extraflags = "-l $lookahead_us";
	full_trace_replay();

	qlogic_disconnect();
	complete();
    }

}


## skew-look

for ($outerloop = $run_skew_look;
     $outerloop > 0;
     $outerloop--) {

    system "make-1000-traces.pl";

    ## adaptec

    for ($innerloop = $run_skew_look_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "skew-look-a";
	config_cheetah();
	config_adaptec();
	prepare();
	adaptec_connect();

	$extraflags = "-k $skew_adaptec_us -l $lookahead_us";
	full_trace_replay();

	adaptec_disconnect();
	complete();
    }

    ## qlogic

    for ($innerloop = $run_skew_look_qlogic;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "skew-look-q";
	config_cheetah();
	config_qlogic();
	prepare();
	qlogic_connect();

	$extraflags = "-k $skew_qlogic_us -l $lookahead_us";
	full_trace_replay();

	qlogic_disconnect();
	complete();
    }

}


########################################################################

## fiftyKrpm lookahead (0->100, mem-sim)

for ($outerloop = $run_fiftyKrpm_lookahead;
     $outerloop > 0;
     $outerloop--) {

    system "./make-100-traces.pl";

    ## adaptec

    for ($innerloop = $run_fiftyKrpm_lookahead_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "fiftyKrpm-lookahead-a";
	config_fiftyKrpm();
	config_adaptec();
	prepare();
	adaptec_connect();

	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 100; $i++) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	adaptec_disconnect();
	complete();
    }
}


## fiftyKrpm-noskew-look

for ($outerloop = $run_fiftyKrpm_noskew_look;
     $outerloop > 0;
     $outerloop--) {

    system "make-1000-traces.pl";

    ## adaptec

    for ($innerloop = $run_fiftyKrpm_noskew_look_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "fiftyKrpm-noskew-look-a";
	config_fiftyKrpm();
	config_adaptec();
	prepare();
	adaptec_connect();

	$extraflags = "-l $lookahead_us";
	full_trace_replay();

	adaptec_disconnect();
	complete();
    }
}


########################################################################

## mems lookahead (0->60, mem-sim)

for ($outerloop = $run_mems_lookahead;
     $outerloop > 0;
     $outerloop--) {

    system "./make-100-traces.pl";

    ## adaptec

    for ($innerloop = $run_mems_lookahead_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "mems-lookahead-a";
	config_mems();
	config_adaptec();
	prepare();
	adaptec_connect();

	$latexcmd = "\\timingEigR";
	$trace = "t-8-r";
	for ($i = 0; $i <= 60; $i++) {
	    $extraflags = "-l $i";
	    trace_replay();
	}

	adaptec_disconnect();
	complete();
    }
}


## mems-noskew-look

for ($outerloop = $run_mems_noskew_look;
     $outerloop > 0;
     $outerloop--) {

    system "make-1000-traces.pl";

    ## adaptec

    for ($innerloop = $run_mems_noskew_look_adaptec;
	 $innerloop > 0;
	 $innerloop--) {

	$runstr = "mems-noskew-look-a";
	config_mems();
	config_adaptec();
	prepare();
	adaptec_connect();

	$extraflags = "-l $lookahead_us";
	full_trace_replay();

	adaptec_disconnect();
	complete();
    }
}


