#!/usr/bin/perl

use English;
use FileHandle;
require "subroutines.pm";


$paper = "compensate";
system "mkdir -p $paper";
system "chmod 777 $paper";

$graph_lookahead_adaptec = 1;
$graph_lookahead_qlogic = 1;
$graph_lookfar_adaptec = 1;
$graph_lookfar_qlogic = 1;
$graph_noskew_nolook_adaptec = 1;
$graph_noskew_nolook_qlogic = 1;
$graph_noskew_look_adaptec = 1;
$graph_noskew_look_qlogic = 1;
$graph_skew_look_adaptec = 1;
$graph_skew_look_qlogic = 1;

#$graph_fiftyKrpm_lookahead_adaptec = 1;
#$graph_fiftyKrpm_noskew_look_adaptec = 1;

#$graph_mems_lookahead_adaptec = 1;
#$graph_mems_noskew_look_adaptec = 1;

## lookahead adaptec

if ($graph_lookahead_adaptec > 0) {

    $runstr = "lookahead-a";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.mem_sim.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookimpactXmemXsimXadaptecXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	my $fofx = "-x + 0.046";
	my $gofx = "0.018";
	my $fofxsym = $fofx;
	$fofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus
	my $gofxsym = $gofx;
	$gofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "f(x) = $fofx\n";
	print GNUFILE "g(x) = $gofx\n";
	print GNUFILE "set ytics 0.01\n";
	print GNUFILE "set xrange [0:0.06]\n";
	print GNUFILE "set yrange [0:0.06]\n";

	print GNUFILE "plot f(x) title \"y = $fofxsym\" lt 3 lw 2, g(x) title \"y = $gofxsym\" lt 2 lw 2, '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}


## lookahead qlogic

if ($graph_lookahead_qlogic > 0) {

    $runstr = "lookahead-q";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.mem_sim.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookimpactXmemXsimXqlogicXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	my $fofx = "-x + 0.046";
	my $gofx = "0.018";
	my $fofxsym = $fofx;
	$fofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus
	my $gofxsym = $gofx;
	$gofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "f(x) = $fofx\n";
	print GNUFILE "g(x) = $gofx\n";
	print GNUFILE "set ytics 0.01\n";
	print GNUFILE "set xrange [0:0.06]\n";
	print GNUFILE "set yrange [0:0.06]\n";

	print GNUFILE "plot f(x) title \"y = $fofxsym\" lt 3 lw 2, g(x) title \"y = $gofxsym\" lt 2 lw 2, '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}


## lookfar adaptec

if ($graph_lookfar_adaptec > 0) {

    $runstr = "lookfar-a";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.hos_mem.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookfarimpactXhosXmemXadaptecXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "set ytics 0.05\n";
	print GNUFILE "set xrange [0:0.40]\n";
	print GNUFILE "set yrange [0:0.40]\n";

	print GNUFILE "plot '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}


## lookfar qlogic

if ($graph_lookfar_qlogic > 0) {

    $runstr = "lookfar-q";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.hos_mem.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookfarimpactXhosXmemXqlogicXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "set ytics 0.05\n";
	print GNUFILE "set xrange [0:0.40]\n";
	print GNUFILE "set yrange [0:0.40]\n";

	print GNUFILE "plot '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}



if ($graph_noskew_nolook_adaptec > 0) {

    $runstr = "noskew-nolook-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_noskew_nolook_qlogic > 0) {

    $runstr = "noskew-nolook-q";
    $hba = "qlogic";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_noskew_look_adaptec > 0) {

    $runstr = "noskew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_noskew_look_qlogic > 0) {

    $runstr = "noskew-look-q";
    $hba = "qlogic";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_skew_look_adaptec > 0) {

    $runstr = "skew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_skew_look_qlogic > 0) {

    $runstr = "skew-look-q";
    $hba = "qlogic";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_fiftyKrpm_noskew_look_adaptec > 0) {

    $runstr = "fiftyKrpm-noskew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_fiftyKrpm_skew_look_adaptec > 0) {

    $runstr = "fiftyKrpm-skew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_mems_noskew_look_adaptec > 0) {

    $runstr = "mems-noskew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


if ($graph_mems_skew_look_adaptec > 0) {

    $runstr = "mems-skew-look-a";
    $hba = "adaptec";

    $which = "mem_sim";
    make_table_and_graph();

    $which = "hos_mem";
    make_table_and_graph();

    $which = "hos_sim";
    make_table_and_graph();
}


########################################################################

sub make_table_and_graph {

    %stampdata = ();
    %minimums = ();
    %maximums = ();

    $whichX = $which;
    $whichX =~ s/_/X/;

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{(.+?)\}\{(\d+)\} % (.+?) $runstr/;
		$stampdata{$2}{latex} = $1;
		$stampdata{$2}{type} = $3;
	    }
	}
	close (SOURCE);
    }

    if (1) {
	foreach my $timestamp (keys %stampdata) {
	    system "mkdir -p timestamps";
	    if (!(-e "timestamps/$timestamp")) {
		print "processing $timestamp\n";
		system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		merge_copy($timestamp);
		process_data_copy($timestamp);
	    }
	}
    }

    # Determine which of the timestamps has the minimum value for $sortval.

    %values = make_statistics_hash(keys %stampdata);

    # Now go through and determine which timestamp contains the minimum
    # value for $sortval across all the runs for a particular latex type.

    foreach my $timestamp (keys %stampdata) {
	my $latex = $stampdata{$timestamp}{latex};

	foreach my $thisw ("mem_sim", "hos_mem", "hos_sim") {
	    foreach my $variable (keys %{$values{$timestamp}{$thisw}}) {
		my $value = $values{$timestamp}{$thisw}{$variable};

		if (!defined $minimums{$latex}{$thisw}{$variable}
		    or $minimums{$latex}{$thisw}{$variable}{value} > $value) {
		    $minimums{$latex}{$thisw}{$variable}{value} = $value;
		    $minimums{$latex}{$thisw}{$variable}{timestamp} = $timestamp;
		}

		if (!defined $maximums{$latex}{$thisw}{$variable}
		    or $maximums{$latex}{$thisw}{$variable}{value} < $value) {
		    $maximums{$latex}{$thisw}{$variable}{value} = $value;
		    $maximums{$latex}{$thisw}{$variable}{timestamp} = $timestamp;
		}
	    }
	}
    }

    ## runstr is in the form of noskew-nolook-a; get rid of the -a.
    $basestr = $runstr;
    chop $basestr;
    chop $basestr;
    $basestr = $basestr . "X$whichX";
    $basestr =~ s/-/X/g;
    $basestr = $basestr . "X$hba";

    ## table

    $texfile = "$paper/$basestr.tex";

    $typestr = $which;
    $typestr =~ s/_/\$\\rightarrow\$/;

    open (TEXFILE, ">$texfile") || die "$texfile";
    print TEXFILE "\\begin{table}\n";
    print TEXFILE "\\centering\n";
    print TEXFILE "\\begin{tabular}{|l|c" . "||c|c"x3 . "|} \\hline\n";
    print TEXFILE "\\DeltaTC\$\\rightarrow\$\\DeltaTR & \\DeltaTRbar & \\eiavg & \\eiavgpct & \\eirms & \\eirmspct & \\earms & \\earmspct \\\\\n";
    print TEXFILE "($typestr) & [ms] & [ms] & [\\%] & [ms] & [\\%] & [ms] & [\\%] \\\\ \\hline\n";

    printf TEXFILE "\\hline \\multicolumn{8}{|l|}{%s} \\\\ \\hline\n", ucfirst($hba);

    foreach $size (2, 8, 64, 128) {
	foreach $rw ("read", "write") {
	    print TEXFILE "\\hline\n";
	    $type    = "t-$size-" . ($rw eq "read" ? "r" : "w");

	    $suffix = "TwoR" if ($size == 2 and $rw eq "read");
	    $suffix = "EigR" if ($size == 8 and $rw eq "read");
	    $suffix = "SxtR" if ($size == 64 and $rw eq "read");
	    $suffix = "HndR" if ($size == 128 and $rw eq "read");
	    $suffix = "TwoW" if ($size == 2 and $rw eq "write");
	    $suffix = "EigW" if ($size == 8 and $rw eq "write");
	    $suffix = "SxtW" if ($size == 64 and $rw eq "write");
	    $suffix = "HndW" if ($size == 128 and $rw eq "write");

	    $latex = "\\timing$suffix";

	    $timestamp = $minimums{$latex}{$which}{EIavg}{timestamp};

	    $output = sprintf "  %s & %.3f & %.3f & %.2f & %.3f & %.2f & %.3f & %.2f \\\\\n",
	    "$size block $rw",
	    $values{$timestamp}{$which}{DeltaTRbar},
	    $values{$timestamp}{$which}{EIavg},
	    $values{$timestamp}{$which}{EIavgpct},
	    $values{$timestamp}{$which}{EIrms},
	    $values{$timestamp}{$which}{EIrmspct},
	    $values{$timestamp}{$which}{EArms},
	    $values{$timestamp}{$which}{EArmspct};

	    $output =~ s/-/--/g;
	    print TEXFILE $output;

	    ## copy fixed graph files

	    system "cp -f timestamps/$timestamp/$timestamp.per.$which.ind.fixed.eps $paper/" . $basestr . "X" . $suffix . "XperXindXfixed.eps";

	    system "cp -f timestamps/$timestamp/$timestamp.per.$which.err.fixed.eps $paper/" . $basestr . "X" . $suffix . "XperXerrXfixed.eps";

	    system "cp -f timestamps/$timestamp/$timestamp.per.$which.pct.fixed.eps $paper/" . $basestr . "X" . $suffix . "XperXpctXfixed.eps";

	    system "cp -f timestamps/$timestamp/$timestamp.agg.$which.fixed.eps $paper/" . $basestr . "X" . $suffix . "XaggXfixed.eps";

	    ## copy auto graph files

	    system "cp -f timestamps/$timestamp/$timestamp.per.$which.ind.auto.eps $paper/" . $basestr . "X" . $suffix . "XperXindXauto.eps";

	}
    }
    print TEXFILE "\\hline\n";
    print TEXFILE "\\end{tabular}\n";
    print TEXFILE "\\$basestr\n";
    print TEXFILE "\\label{$basestr}\n";
    print TEXFILE "\\end{table}\n";
    close (TEXFILE);

    ## fixed graphs

    foreach $suffix ("TwoR", "TwoW", "EigR", "EigW", "SxtR", "SxtW", "HndR", "HndW") {

	$newbase = "$basestr" . "X$suffix";

	$texfile = "$paper/$newbase" . "Xfixed.tex";

	open (TEXFILE, ">$texfile") || die "$texfile";

	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\compensateEpsArrayFixed{$newbase}\n";
	print TEXFILE "\\$newbase" . "Xfixed\n";
	print TEXFILE "\\label{$newbase" . "Xfixed}\n";
	print TEXFILE "\\end{figure}\n";

	close (TEXFILE);
    }

    ## auto graphs

    $newbase = "$basestr";
    $texfile = "$paper/$newbase" . "Xauto.tex";

    open (TEXFILE, ">$texfile") || die "$texfile";

    print TEXFILE "\\begin{figure}\n";
    print TEXFILE "\\compensateEpsArrayAutoBoth{$basestr}\n";
    print TEXFILE "\\$newbase" . "Xauto\n";
    print TEXFILE "\\label{$newbase" . "Xauto}\n";
    print TEXFILE "\\end{figure}\n";

    close (TEXFILE);

#    foreach $suffix ("TwoR", "TwoW", "EigR", "EigW", "SxtR", "SxtW", "HndR", "HndW") {
#
#	$newbase = "$basestr" . "X$suffix";
#
#	$texfile = "$paper/$newbase" . "Xauto.tex";
#
#	$typestr = $which;
#	$typestr =~ s/_/\$\\rightarrow\$/;
#
#	open (TEXFILE, ">$texfile") || die "$tablefile";
#
#	print TEXFILE "\\begin{figure}\n";
#	print TEXFILE "\\centering\n";
#	print TEXFILE "\\compensateEpsArrayAuto{$newbase}\n";
#	print TEXFILE "\\$newbase" . "Xauto\n";
#	print TEXFILE "\\label{$newbase" . "Xauto}\n";
#	print TEXFILE "\\end{figure}\n";
#
#	close (TEXFILE);
#    }

}


if ($graph_fiftyKrpm_lookahead_adaptec > 0) {

    $runstr = "fiftyKrpm-lookahead-a";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.mem_sim.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookfiftyKXmemXsimXadaptecXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	my $fofx = "-x + 0.075";
	my $gofx = "0.018";
	my $fofxsym = $fofx;
	$fofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus
	my $gofxsym = $gofx;
	$gofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "f(x) = $fofx\n";
	print GNUFILE "g(x) = $gofx\n";
	print GNUFILE "set ytics 0.01\n";
	print GNUFILE "set xrange [0:0.1]\n";
	print GNUFILE "set yrange [0:0.1]\n";

	print GNUFILE "plot f(x) title \"y = $fofxsym\" lt 3 lw 2, g(x) title \"y = $gofxsym\" lt 2 lw 2, '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}


if ($graph_mems_lookahead_adaptec > 0) {

    $runstr = "mems-lookahead-a";

    # Allstamps is a hash indexed by all the lookahead values used.
    # We want to read in all the source files, and identify the minimum
    # value of EIavg calculated across all the runs.

    %allstamps = ();

    foreach $runid (`ls -1 runs/$runstr`) {
	chomp $runid;

	my $sourcefile = "runs/$runstr/$runid/source";

	open (SOURCE, "$sourcefile") || die "$sourcefile";
	while (<SOURCE>) {
	    if (!(/^%/)) {  # if line doesn't begin with %
		/.+?\{.+?\}\{(\d+)\} % (.+?) $runstr -l (\d+)/;
		push @{$allstamps{$3}{stamps}}, $1;
	    }
	}
	close (SOURCE);
    }

    # Process the data in the source files, if necessary.

    if (1) {
	foreach my $lookahead (keys %allstamps) {
	    foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
		system "mkdir -p timestamps";
		if (!(-e "timestamps/$timestamp")) {
		    print "processing $timestamp\n";
		    system "cp -fr $results_dir/$timestamp timestamps/$timestamp";
		    merge_copy($timestamp);
		    process_data_copy($timestamp);
		}
	    }
	}
    }

    # Now go through and calculate the minimum value for EIavg across all
    # the runs.

    foreach my $lookahead (keys %allstamps) {
	foreach my $timestamp (@{$allstamps{$lookahead}{stamps}}) {
	    foreach my $peragg ("per", "agg") {
		my $statsfile = "timestamps/$timestamp/$timestamp.$peragg.mem_sim.stats";

		open (STATS, "$statsfile") or die "Cannot open $statsfile\n";
		while (<STATS>) {
		    my ($variable, $equals, $value, $units) = split;
		    if (!defined $allstamps{$lookahead}{$variable}
			or $allstamps{$lookahead}{$variable} > $value) {
			$allstamps{$lookahead}{$variable} = $value;
		    } 
		}
		close (STATS);
	    }
	}
    }

    # Finally, print out the list of the minimum values for EIavg.

#    foreach my $what ("EIavg", "EIrms", "EArms") {
    foreach $what ("EIavg") {
	my $whatyaxis;
	$whatyaxis = "EI_{avg}" if ($what eq "EIavg");
	$whatyaxis = "EI_{rms}" if ($what eq "EIrms");
	$whatyaxis = "EA_{rms}" if ($what eq "EArms");

	$basestr = "lookmemsXmemXsimXadaptecXEIavg";
	$datfile = "$paper/$basestr.dat";
	$gnufile = "$paper/$basestr.gnu";
	$epsfile = "$paper/$basestr.eps";
	$texfile = "$paper/$basestr.tex";

	open (DATFILE, "|sort -g >$datfile") or die ("Cannot open $datfile\n");
	foreach my $lookahead (keys %allstamps) {
	    printf DATFILE "%.3f %.3f\n", 
	    $lookahead / 1000.0, 
	    $allstamps{$lookahead}{$what};
	}
	close (DATFILE);

	open (GNUFILE, ">$gnufile") or die "Cannot open $gnufile\n";

	my $fofx = "-x + 0.042";
	my $gofx = "0.016";
	my $fofxsym = $fofx;
	$fofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus
	my $gofxsym = $gofx;
	$gofxsym =~ s/-/{\/Symbol -}/g;  # convert minus to long-minus

	print GNUFILE "set term postscript enhanced color \"Helvetica\" 24\n";
	print GNUFILE "set title \"{/Helvetica-Bold Impact of current-time lookahead on observed error}\"\n";
	print GNUFILE "set xlabel \"T_{lookahead}, units=ms\\n\"\n";
	print GNUFILE "set ylabel \"E_{2{/Symbol \\256}3}: $whatyaxis, units=ms\"\n";
	print GNUFILE "f(x) = $fofx\n";
	print GNUFILE "g(x) = $gofx\n";
	print GNUFILE "set ytics 0.01\n";
	print GNUFILE "set xrange [0:0.06]\n";
	print GNUFILE "set yrange [0:0.06]\n";

	print GNUFILE "plot f(x) title \"y = $fofxsym\" lt 3 lw 2, g(x) title \"y = $gofxsym\" lt 2 lw 2, '$datfile' title '4 KB read trace' with linespoints lt -1 lw 2\n";
	print GNUFILE "quit\n";

	close(GNUFILE);

	system "gnuplot $gnufile > $epsfile";

	open (TEXFILE, ">$texfile") or die "$outfile\n";
	print TEXFILE "\\begin{figure}\n";
	print TEXFILE "\\centering\n";
	print TEXFILE "\\resizebox{4.5in}{!}{\\rotatebox{-90}{\\epsfig{file=compensate/$basestr}}}\n";
	print TEXFILE "\\$basestr\n";
	print TEXFILE "\\label{$basestr}\n";
	print TEXFILE "\\end{figure}\n";
	close (TEXFILE);
    }
}


