#!/usr/bin/perl

## Okay, what exactly am I doing here?  I want to start the emulator,
## run a specific workload (to be written later), then analyze the
## results and append them to a file for later graphing.

use English;
use FileHandle;
require "subroutines.pm";

my $runstr = "skew-look";

my $runid = `date +%s`; chomp ($runid);
my $rundir = "data/$runid";
system "mkdir data/$runid";

my @stamps = ();
my $stampsfile = "$rundir/stamps";
my $sourcefile = "$rundir/source";

open (STAMPS, ">$stampsfile") || die "$stampsfile";
open (SOURCE, ">$sourcefile") || die "$sourcefile";

autoflush STAMPS 1;
autoflush SOURCE 1;

open_log($runid);

print SOURCE "% $runstr\n";
print SOURCE "% Begin: " . `date`;

run_trace("\\timingTwoR", "t-2-r");
run_trace("\\timingTwoW", "t-2-w");
run_trace("\\timingEigR", "t-8-r");
run_trace("\\timingEigW", "t-8-w");
run_trace("\\timingSxtR", "t-64-r");
run_trace("\\timingSxtW", "t-64-w");
run_trace("\\timingHndR", "t-128-r");
run_trace("\\timingHndW", "t-128-w");

print SOURCE "% End: " . `date`;

close_log();
close (SOURCE);
close (STAMPS);

foreach my $timestamp (@stamps) {
    merge($timestamp);
    process_data($timestamp);
}

make_tables($runid);
make_summary($runid);

system "ln -s ../$rundir $runstr";
system "rm lastrun; ln -s $rundir lastrun";


sub run_trace {
    my ($latexcmd, $trace) = @_;

    die "Stop file exists!\n" if (-e "stop");

    if (1) {
	my $timestamp = `date +%s`; chomp($timestamp);
	print STAMPS "$timestamp\n";
	push @stamps, $timestamp;
	my $newcmd = $latexcmd . "A";
	# 0.099 + 0.015 = 0.114
	# extra lookahead of 0.035
	$extraflags = "-k 114 -l 149 -y 71687370";
	print SOURCE "\\newcommand{$newcmd}{$timestamp} % $trace adaptec $extraflags\n";
	execute_trace($timestamp, $trace, 0);
	sleep(10);
    }

    die "Stop file exists!\n" if (-e "stop");

    if (0) {
	my $timestamp = `date +%s`; chomp($timestamp);
	print STAMPS "$timestamp\n";
	push @stamps, $timestamp;
	my $newcmd = $latexcmd . "Q";
	# 0.282 + 0.015 = 0.297
	# extra lookahead of 0.035
	$extraflags = "-k 297 -l 332 -y 71687370";
	print SOURCE "\\newcommand{$newcmd}{$timestamp} % $trace qlogic $extraflags\n";
	execute_trace($timestamp, $trace, 1);
	sleep(10);
    }
}


########################################################################

sub execute_trace {
    my ($timestamp, $synthtrace, $use_qlogic) = @_;

    $use_qlogic ? qlogic_start($timestamp) : adaptec_start($timestamp);

    #ssh_cmd_checkresult("$host_name", "(cd $main_dir/microgenerate ; ./microgenerate -g 2)", 0);  ## -g 2 means it's the third disk on the system -- should be constant

    ssh_cmd_checkresult("$host_name", "(cd $main_dir/timing ; $main_dir/src/disksim/dixtrac/dxreplay -s -m $synthtrace /dev/sg2)", 0);

    $use_qlogic ? qlogic_stop($timestamp) : adaptec_stop($timestamp);
}

