#!/usr/bin/perl

## Okay, what exactly am I doing here?  I want to start the emulator,
## run a specific workload (to be written later), then analyze the
## results and append them to a file for later graphing.

use English;
use FileHandle;
require "subroutines.pm";

my $runstr = "postmark";

my $runid = `date +%s`; chomp ($runid);
my $rundir = "data/$runid";
print "$rundir\n";
system "mkdir $rundir";

system "ln -s ../$rundir $runstr";
system "rm lastrun; ln -s $rundir lastrun";

# $extraflags = "-k 114 -l 149 -z 71687370 -y 1837056 -x /usr/disk-images/896mb";
# $extraflags = "-z 1837056 -y 1837056 -x /usr/disk-images/896mb";
$extraflags = "-k 58 -l 86 -z 71687370 -y 1837056 -x /usr/disk-images/896mb";
# $extraflags = "-k 118 -l 149 -z 71687370 -y 1837056 -x /usr/disk-images/896mb";

my $sourcefile = "$rundir/source";

open (SOURCE, ">$sourcefile") || die "$sourcefile";
autoflush SOURCE 1;
open_log("$rundir/runlog.$runid");

print SOURCE "% $runstr\n";
print SOURCE "% Begin: " . `date`;

#runit_postmark_localdisk();
runit_postmark_adaptec_cheetah();
#runit_postmark_adaptec_mems();
#runit_postmark_adaptec_immediate();

print SOURCE "% End: " . `date`;

close_log();
close (SOURCE);



sub runit_postmark_localdisk {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark local localdisk";
    log_entry("Running: $descr $timestamp\n");

    mkdir "$results_dir/$timestamp", 0777 or die "Cannot mkdir $timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    $host_address_host_no = 2;
    $host_address_channel = 0;
    $host_address_target  = 2;
    $host_address_lun     = 0;

    host_dxcachectl_sdc();

    ssh_cmd_checkresult("root\@$host_name", "dd if=/tmp/896mb of=/dev/sdc bs=1048576", 0);
    ssh_cmd_checkresult("root\@$host_name", "mount -osync /dev/sdc1 /mnt/memulator", 0);

    host_insmod_hosttimes($timestamp);
    do_postmark($timestamp, "$descr");
    host_rmmod_hosttimes($timestamp);
    ssh_cmd_checkresult("root\@$host_name", "umount /mnt/memulator");

    merge_hos_only($timestamp);
}


sub runit_postmark_qlogic_mems {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark qlogic mems";
    log_entry("Running: $descr $timestamp\n");
    system "ln -sf mems.g2.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    my_start_qlogic($timestamp);
    do_postmark($timestamp, "$descr");
    my_stop_qlogic();

    merge($timestamp);
    process_data($timestamp);
}


sub runit_postmark_adaptec_mems {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark adaptec mems";
    log_entry("Running: $descr $timestamp\n");
    system "ln -sf mems.g2.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    my_start_adaptec($timestamp, 1);
    do_postmark($timestamp, "$descr");
    my_stop_adaptec();

    merge($timestamp);
    process_data($timestamp);
}


sub runit_postmark_qlogic_cheetah {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark qlogic cheetah";
    log_entry("Running: $descr $timestamp\n");
    system "ln -sf cheetah.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    my_start_qlogic($timestamp);
    sleep 1;
    do_postmark($timestamp, "$descr");
    sleep 1;
    my_stop_qlogic();

    merge($timestamp);
    process_data($timestamp);
}

sub runit_postmark_adaptec_cheetah {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark adaptec cheetah";
    log_entry("Running: $descr $timestamp\n");
    system "ln -sf cheetah.parv /home/griffin2/proj/memulator/src/memulator/parfile";
    my_start_adaptec($timestamp, 1);
    sleep 1;
    do_postmark($timestamp, "$descr");
    sleep 1;
    my_stop_adaptec();

    merge($timestamp);
    process_data($timestamp);
}

sub runit_postmark_adaptec_immediate {
    my $timestamp = `date +%s`; chomp($timestamp);
    my $descr = "postmark adaptec immediate";
    log_entry("Running: $descr $timestamp\n");
    my_start_adaptec($timestamp, 0);
    sleep 1;
    do_postmark($timestamp, "$descr");
    sleep 1;
    my_stop_adaptec();
}



sub do_postmark {
    my ($timestamp, $descr) = @_;

#    @ssh_output = ();
    ssh_cmd("root\@$host_name", "cp /home/griffin2/workbench/mems-exps/bin/* /mnt/memulator", 0);
    ssh_cmd("root\@$host_name", "(cd /mnt/memulator ; echo 'run' | ./postmark)", 0);
#    print SOURCE "$timestamp $descr $ssh_lastline\n";
}


sub my_start_adaptec {
    my ($timestamp, $useclock) = @_;

    mkdir "$results_dir/$timestamp", 0777 or die "Cannot mkdir $timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    config_adaptec();

    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_add_single_device();
    sleep 1;
    host_mount_memulator_sync();
    sleep 1;
    targ_stop_memulator();

    if ($useclock == 1) {
	targ_start_memulator_times($timestamp);
    } else {
	targ_start_memulator_notimes($timestamp);
    }
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub my_stop_adaptec {
    config_adaptec();

    host_umount_memulator();
    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_remove_single_device();
    targ_stop_memulator();
}



sub my_start_qlogic {
    my ($timestamp) = @_;

    mkdir "$results_dir/$timestamp", 0777 or die "Cannot mkdir $timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    config_qlogic();

    host_rmmod_qla2200();
    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_insmod_qla2200();
    host_mount_memulator_sync();
    targ_stop_memulator();

    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub my_stop_qlogic {
    config_qlogic();

    host_umount_memulator();
    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_rmmod_qla2200();
    targ_stop_memulator();
}


die;



my @stamps = ();
my $stampsfile = "$rundir/stamps";
my $sourcefile = "$rundir/source";

open (STAMPS, ">$stampsfile") || die "$stampsfile";
open (SOURCE, ">$sourcefile") || die "$sourcefile";

autoflush STAMPS 1;
autoflush SOURCE 1;

open_log($runid);

print SOURCE "% $runstr\n";
print SOURCE "% Begin: " . `date`;

run_trace("\\timingTwoR", "t-2-r");
run_trace("\\timingTwoW", "t-2-w");
run_trace("\\timingEigR", "t-8-r");
run_trace("\\timingEigW", "t-8-w");
run_trace("\\timingSxtR", "t-64-r");
run_trace("\\timingSxtW", "t-64-w");
run_trace("\\timingHndR", "t-128-r");
run_trace("\\timingHndW", "t-128-w");

print SOURCE "% End: " . `date`;

close_log();
close (SOURCE);
close (STAMPS);

foreach my $timestamp (@stamps) {
    merge($timestamp);
    process_data($timestamp);
}

make_tables($runid);
make_summary($runid);

system "ln -s ../$rundir $runstr";
system "rm lastrun; ln -s $rundir lastrun";


sub run_trace {
    my ($latexcmd, $trace) = @_;

    die "Stop file exists!\n" if (-e "stop");

    if (0) {
	my $timestamp = `date +%s`; chomp($timestamp);
	print STAMPS "$timestamp\n";
	push @stamps, $timestamp;
	my $newcmd = $latexcmd . "A";
	# 0.099 + 0.015 = 0.114
	# extra lookahead of 0.035
	$extraflags = "-k 114 -l 149";
	print SOURCE "\\newcommand{$newcmd}{$timestamp} % $trace adaptec $extraflags\n";
	execute_trace($timestamp, $trace, 0);
	sleep(10);
    }

    die "Stop file exists!\n" if (-e "stop");

    if (1) {
	my $timestamp = `date +%s`; chomp($timestamp);
	print STAMPS "$timestamp\n";
	push @stamps, $timestamp;
	my $newcmd = $latexcmd . "Q";
	# 0.282 + 0.015 = 0.297
	# extra lookahead of 0.035
	$extraflags = "-k 297 -l 332";
	print SOURCE "\\newcommand{$newcmd}{$timestamp} % $trace qlogic $extraflags\n";
	execute_trace($timestamp, $trace, 1);
	sleep(10);
    }
}


########################################################################

sub execute_trace {
    my ($timestamp, $synthtrace, $use_qlogic) = @_;

    $use_qlogic ? qlogic_start($timestamp) : adaptec_start($timestamp);

    #ssh_cmd_checkresult("$host_name", "(cd $main_dir/microgenerate ; ./microgenerate -g 2)", 0);  ## -g 2 means it's the third disk on the system -- should be constant

    ssh_cmd_checkresult("$host_name", "(cd $main_dir/timing ; $main_dir/src/disksim/dixtrac/dxreplay -s -m $synthtrace /dev/sg2)", 0);

    $use_qlogic ? qlogic_stop($timestamp) : adaptec_stop($timestamp);
}

