
#$devsize = 2097152;
$devsize = 600000;

$host_name = "nc93.pdl.cmu.edu";  # host machine
$targ_name = "nc94.pdl.cmu.edu";  # target (emulation) machine

# $host_cycles = 1993907171.0;
# $targ_cycles = 1993868287.0;

$host_cycles = 1993908205.0;  # four-day calibrated average
$targ_cycles = 1993868931.0;  # four-day calibrated average

$main_dir = "/home/griffin2/workbench";
$results_dir = "$main_dir/results";

########################################################################

sub open_log {
    my ($logfile) = @_;
    open(LOG, ">>$logfile") or die "Cannot open log file $logfile\n";
    autoflush LOG 1;
}

sub log_entry {
    my ($msg) = @_;
    print LOG "$msg";
    print "$msg";
}

sub close_log {
    close(LOG);
}

sub panic {
    my ($msg) = @_;
    print "###############################################################\n";
    print "########################### FAILURE ###########################\n";
    print "###############################################################\n";
    log_entry($msg);
    die "$msg";
}


sub ssh_cmd {
    my ($destination,$command) = @_;

    log_entry("ssh $destination $command\n");

    # fork a process for the ssh
    my $pid = open(KID, "-|");
    panic "unable to fork" unless defined $pid;

    if ($pid == 0) {  # child process
	open STDERR,'>&STDOUT';
	@cmd = ("ssh","$destination","$command");
	exec(@cmd);
	# NOTREACHED
    } else {          # parent process
	while ($output = <KID>) {
	    log_entry("ssh: $output");
	    $ssh_lastline = $output;
	}
	close KID;
	return $?;
    }
}

sub ssh_cmd_checkresult {
    my ($destination,$command,$expected_result) = @_;

    my $result = ssh_cmd($destination,$command);

    if ($result != $expected_result) {
	panic("bad ssh result ($result != $expected_result)");
    }
}

########################################################################

sub targ_start_memulator_times {
    my ($timestamp) = @_;
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -b $targ_memulator_store -c -f $results_dir/$timestamp/$timestamp $targ_memulator_address &) < /dev/null >& /dev/null");
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -b /dev/da1 -c -f $results_dir/$timestamp/$timestamp -z 71687370 $extraflags $targ_memulator_address &) < /dev/null >& $results_dir/$timestamp/$timestamp.mem.log");
    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -c -f $results_dir/$timestamp/$timestamp $extraflags $targ_memulator_address &) < /dev/null >& $results_dir/$timestamp/$timestamp.mem.log");
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -f $results_dir/$timestamp/$timestamp $extraflags $targ_memulator_address &) < /dev/null >& $results_dir/$timestamp/$timestamp.mem.log");
    sleep 60;
}


sub targ_start_memulator_notimes {
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -b $targ_memulator_store $targ_memulator_address &) < /dev/null >& /dev/null");
    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator $extraflags $targ_memulator_address &) < /dev/null >& /dev/null");
    sleep 60;
}


sub targ_verify_memulator {
    ssh_cmd_checkresult("$targ_name", "ps ax | grep memulator | grep -v grep", 0);
}


sub host_insmod_qla2200 {
    ssh_cmd_checkresult("root\@$host_name", "insmod qla2200", 0);
}


sub host_add_single_device {
    ssh_cmd_checkresult("root\@$host_name", "echo 'scsi add-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi", 0);
}


sub host_insmod_hosttimes {
    my ($timestamp) = @_;
    ssh_cmd_checkresult("root\@$host_name", "insmod /home/griffin2/workbench/hosttimes/hosttimes.o filename=$results_dir/$timestamp/$timestamp.hos host_no=$host_address_host_no channel=$host_address_channel target=$host_address_target lun=$host_address_lun", 0);
}


sub host_mount_memulator_sync {
    ssh_cmd_checkresult("root\@$host_name", "mount -osync /dev/sdd1 /mnt/memulator", 0);
}


sub host_umount_memulator {
    ssh_cmd_checkresult("root\@$host_name", "umount /mnt/memulator", 0);
}


sub host_dxcachectl_sdc {
    ssh_cmd_checkresult("root\@$host_name", "/home/griffin2/workbench/mems-exps/bin/dxcachectl -c OFF /dev/sdc", 0);
}


########################################################################

sub host_rmmod_hosttimes {
    ssh_cmd("root\@$host_name", "rmmod hosttimes");
}


sub host_rmmod_qla2200 {
    ssh_cmd("root\@$host_name", "rmmod qla2200");
}


sub host_remove_single_device {
#    ssh_cmd_checkresult("root\@$host_name", "echo 'scsi remove-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi", 256);
    ssh_cmd("root\@$host_name", "echo 'scsi remove-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi");
}


sub targ_stop_memulator {
    ssh_cmd_checkresult("root\@$targ_name", "killall -1 memulator", 0);
}

########################################################################

sub config_qlogic {
    $targ_memulator_store = "/dev/da1";  # backing store for emulated disk
    $targ_memulator_address = "1:94:0";  # bus:target:lun for emualted disk

    $host_address_host_no = 3;
    $host_address_channel = 0;
    $host_address_target  = 0;
    $host_address_lun     = 0;
}


sub config_adaptec {
    $targ_memulator_store = "/dev/da1";  # backing store for emulated disk
    $targ_memulator_address = "2:4:0";   # bus:target:lun for emualted disk

    $host_address_host_no = 1;
    $host_address_channel = 0;
    $host_address_target  = 4;
    $host_address_lun     = 0;
}


sub qlogic_start {
    my ($timestamp) = @_;

    mkdir "$results_dir/$timestamp", 0777 or die "Cannot mkdir $timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    config_qlogic();

    host_rmmod_qla2200();
    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_insmod_qla2200();
    targ_stop_memulator();

    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub qlogic_stop {
    config_qlogic();

    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_rmmod_qla2200();
    targ_stop_memulator();
}


sub adaptec_start {
    my ($timestamp) = @_;

    mkdir "$results_dir/$timestamp", 0777 or die "Cannot mkdir $timestamp\n";
    chmod 0777, "$results_dir/$timestamp";

    config_adaptec();

    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_add_single_device();
    targ_stop_memulator();

    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub adaptec_stop {
    config_adaptec();

    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_remove_single_device();
    targ_stop_memulator();
}

########################################################################

sub process_data {
    my ($timestamp) = @_;
    my $prefix = "$results_dir/$timestamp/$timestamp";

    my @requests = process_merge_file($prefix);
    process_print_datfile($prefix, \@requests);

    process_per_request($prefix, \@requests);
    process_aggregate($prefix, \@requests);
}


sub process_merge_file {
    my ($prefix) = @_;

    my $fh_name = "$prefix.merge";
    open(FH, "$fh_name") || panic "Cannot open $fh_name";

    my @requests = ();

    my $hos_arrive_first;
    my $mem_arrive_first;
    my $hos_complete_first;
    my $mem_complete_first;

    while (<FH>) {
	my ($num, 
	    $offset, $length, 
	    $opcode, 
	    $hos_arrive, $hos_complete, 
	    $mem_arrive, $mem_complete, 
	    $simtime) = split;

	$hos_arrive_first = $hos_arrive unless defined $hos_arrive_first;
	$mem_arrive_first = $mem_arrive unless defined $mem_arrive_first;
	$hos_complete_first = $hos_complete unless defined $hos_complete_first;
	$mem_complete_first = $mem_complete unless defined $mem_complete_first;

	my $hos_ms = ($hos_complete - $hos_arrive) / $host_cycles * 1000.0;
	my $mem_ms = ($mem_complete - $mem_arrive) / $targ_cycles * 1000.0;
	my $sim_ms = $simtime;

	my $hos_sim_err_ms  = $hos_ms - $sim_ms;
	my $hos_sim_err_pct = $hos_sim_err_ms / $sim_ms * 100.0;

	my $mem_sim_err_ms  = $mem_ms - $sim_ms;
	my $mem_sim_err_pct = $mem_sim_err_ms / $sim_ms * 100.0;

	my $hos_mem_err_ms  = $hos_ms - $mem_ms;
	my $hos_mem_err_pct = $hos_mem_err_ms / $mem_ms * 100.0;

	my $hos_arrive_time_ms = 
	    ($hos_arrive - $hos_arrive_first) / $host_cycles * 1000.0;
	my $mem_arrive_time_ms =
	    ($mem_arrive - $mem_arrive_first) / $targ_cycles * 1000.0;
	my $hos_complete_time_ms = 
	    ($hos_complete - $hos_complete_first) / $host_cycles * 1000.0;
	my $mem_complete_time_ms =
	    ($mem_complete - $mem_complete_first) / $targ_cycles * 1000.0;

	push @requests, {
	    hos_ms => $hos_ms,
	    mem_ms => $mem_ms,
	    sim_ms => $sim_ms,
	    hos_sim_err_ms  => $hos_sim_err_ms,
	    hos_sim_err_pct => $hos_sim_err_pct,
	    mem_sim_err_ms  => $mem_sim_err_ms,
	    mem_sim_err_pct => $mem_sim_err_pct,
	    hos_mem_err_ms  => $hos_mem_err_ms,
	    hos_mem_err_pct => $hos_mem_err_pct,
	    hos_arrive_time_ms => $hos_arrive_time_ms,
	    mem_arrive_time_ms => $mem_arrive_time_ms,
	    hos_complete_time_ms => $hos_complete_time_ms,
	    mem_complete_time_ms => $mem_complete_time_ms,
	};
    }

    close(FH);

    return @requests;
}


sub process_print_datfile {
    my ($prefix, $requests) = @_;

    my $out_name = "$prefix.dat";
    open(OUT, ">$out_name") || panic "Cannot open $out_name";

    for (my $i = 0; $i < @$requests; $i++) {
	printf OUT 
	    "%d\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f%%\t%7.3f\t%7.3f%%\t%7.3f\t%7.3f%%\t%7.3f\t%7.3f\n",
	    $i + 1,
	    @$requests[$i]->{hos_ms},
	    @$requests[$i]->{mem_ms},
	    @$requests[$i]->{sim_ms},
	    @$requests[$i]->{hos_sim_err_ms },
	    @$requests[$i]->{hos_sim_err_pct},
	    @$requests[$i]->{mem_sim_err_ms },
	    @$requests[$i]->{mem_sim_err_pct},
	    @$requests[$i]->{hos_mem_err_ms },
	    @$requests[$i]->{hos_mem_err_pct},
	    @$requests[$i]->{mem_arrive_time_ms} - @$requests[$i]->{hos_arrive_time_ms},
	    @$requests[$i]->{mem_complete_time_ms} - @$requests[$i]->{hos_complete_time_ms};
    }

    close(OUT);
}


sub process_per_request {
    my ($prefix, $requests) = @_;

    process_per_request_stats_histograms($prefix, \@$requests, "hos", "sim");
    process_per_request_stats_histograms($prefix, \@$requests, "mem", "sim");
    process_per_request_stats_histograms($prefix, \@$requests, "hos", "mem");
}


sub process_per_request_stats_histograms {
    my ($prefix, $requests, $left, $right) = @_;

    my $which = "$left\_$right";

    my $N = @$requests;

    my $DeltaTRbar = process_avg(\@$requests, "$right\_ms");
    my $DeltaTRmin = process_min(\@$requests, "$right\_ms");
    my $DeltaTRmax = process_max(\@$requests, "$right\_ms");

    my $DeltaTCbar = process_avg(\@$requests, "$left\_ms");
    my $DeltaTCmin = process_min(\@$requests, "$left\_ms");
    my $DeltaTCmax = process_max(\@$requests, "$left\_ms");

    my $EIavg = process_avg(\@$requests, "$left\_$right\_err_ms");
    my $EIavgpct = $EIavg / $DeltaTRbar * 100.0;

    my $EImin = process_min(\@$requests, "$left\_$right\_err_ms");
    my $EIminpct = $EImin / $DeltaTRbar * 100.0;

    my $EImax = process_max(\@$requests, "$left\_$right\_err_ms");
    my $EImaxpct = $EImax / $DeltaTRbar * 100.0;

    my $EIrms = process_rms(\@$requests, "$left\_$right\_err_ms");
    my $EIrmspct = $EIrms / $DeltaTRbar * 100.0;

    my $fh_name = "$prefix.per.$left\_$right.stats";
    open(FH, ">$fh_name" || panic "Cannot open $fh_name\n");

    printf FH "N          = %d requests\n", $N;
    printf FH "DeltaTRbar = %f ms\n", $DeltaTRbar;
    printf FH "DeltaTRmin = %f ms\n", $DeltaTRmin;
    printf FH "DeltaTRmax = %f ms\n", $DeltaTRmax;
    printf FH "DeltaTCbar = %f ms\n", $DeltaTCbar;
    printf FH "DeltaTCmin = %f ms\n", $DeltaTCmin;
    printf FH "DeltaTCmax = %f ms\n", $DeltaTCmax;
    printf FH "EIavg      = %f ms\n", $EIavg;
    printf FH "EIavgpct   = %f %%\n", $EIavgpct;
    printf FH "EImin      = %f ms\n", $EImin;
    printf FH "EIminpct   = %f %%\n", $EIminpct;
    printf FH "EImax      = %f ms\n", $EImax;
    printf FH "EImaxpct   = %f %%\n", $EImaxpct;
    printf FH "EIrms      = %f ms\n", $EIrms;
    printf FH "EIrmspct   = %f %%\n", $EIrmspct;

    close(FH);

    ### Individual error graph

    my $datfile = "$prefix.dat";
    my $epsfile = "$prefix.per.$which.ind.eps";

    my $datcolumn;
    if ($which eq "hos_sim") { $datcolumn = 5; }
    elsif ($which eq "mem_sim") { $datcolumn = 7; }
    elsif ($which eq "hos_mem") { $datcolumn = 9; }
    else { die "Unable to select column for $which\n"; }

    open(GNU, "|gnuplot > $epsfile") || panic "Failed $epsfile\n";

    # Note: the \\305's produce the overbar for the DeltaTRbar

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Individual request errors}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EI_{avg}=%.3f ms, EI_{rms}=%.3f ms\"\n", $DeltaTRbar, $EIavg, $EIrms;
    printf GNU "set xtics %d\n", $N/5;
    print GNU "set xlabel 'Request number'\n";
    print GNU "set ylabel \"{/Symbol D}T_C(n) {/Symbol - D}T_R(n), units=ms\"\n";
    print GNU "plot '$datfile' using 1:$datcolumn title '' with points lt -1 pt 7 ps 0.5\n";
    print GNU "quit\n";

    close(GNU);

    ### Error histogram

    my $datfile = "$prefix.per.$which.err.dat";
    my $epsfile_lg = "$prefix.per.$which.err.lg.eps";
    my $epsfile_sm = "$prefix.per.$which.err.sm.eps";
    my $bucketsize = 0.02;
    my %buckets = ();

    open(DAT, "|sort -g >$datfile") || panic "Cannot open $datfile\n";

    my $xrange_min = process_min(\@$requests, "$which\_err_ms");
    $xrange_min = 0.0 if ($xrange_min > 0);

    foreach my $request (@$requests) {
	my $bucket;
	my $val = $request->{"$which\_err_ms"};

	$bucket = int($val / $bucketsize) * $bucketsize;
	if ($val >= 0.0) {
	    $bucket += ($bucketsize / 2.0);
	} else {
	    $bucket -= ($bucketsize / 2.0);
	}

	$buckets{$bucket}++;
    }

    foreach my $key (keys %buckets) {
	print DAT "$key $buckets{$key}\n";
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile_lg") || panic "Failed $epsfile_lg\n";

    # Note: the \\305's produce the overbar for the DeltaTRbar

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Individual request error histogram}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EI_{avg}=%.3f ms, EI_{rms}=%.3f ms\"\n", $DeltaTRbar, $EIavg, $EIrms;
    print GNU "set xrange [$xrange_min:]\n";
    print GNU "set yrange [0:$N]\n";
    print GNU "set xlabel \"{/Symbol D}T_C(n) {/Symbol - D}T_R(n), units=ms\"\n";
    print GNU "set ylabel 'Number of requests (N=$N)'\n";
    print GNU "set boxwidth $bucketsize\n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3\n";
    print GNU "quit\n";

    close(GNU);

    ## small version

    open(GNU, "|gnuplot > $epsfile_sm") || panic "Failed $epsfile_sm\n";

    # Note: the \\305's produce the overbar for the DeltaTRbar

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Individual request error histogram}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EI_{avg}=%.3f ms, EI_{rms}=%.3f ms\"\n", $DeltaTRbar, $EIavg, $EIrms;
    print GNU "set xtics 0.1\n";
    print GNU "set xrange [-0.5:0.5]\n";
    print GNU "set yrange [0:$N]\n";
    print GNU "set xlabel \"{/Symbol D}T_C(n) {/Symbol - D}T_R(n), units=ms\"\n";
    print GNU "set ylabel 'Number of requests (N=$N)'\n";
    print GNU "set boxwidth $bucketsize\n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3\n";
    print GNU "quit\n";

    close(GNU);

    ### Percentage error histogram

    my $datfile = "$prefix.per.$which.pct.dat";
    my $epsfile_lg = "$prefix.per.$which.pct.lg.eps";
    my $epsfile_sm = "$prefix.per.$which.pct.sm.eps";
    my $bucketsize = 1.0;
    my %buckets = ();

    open(DAT, "|sort -g >$datfile") || panic "Cannot open $datfile\n";

    my $xrange_min = process_min(\@$requests, "$which\_err_pct");
    $xrange_min = 0.0 if ($xrange_min > 0);

    foreach my $request (@$requests) {
	my $bucket;
	my $val = $request->{"$which\_err_pct"};

	$bucket = int($val / $bucketsize) * $bucketsize;
	if ($val >= 0.0) {
	    $bucket += ($bucketsize / 2.0);
	} else {
	    $bucket -= ($bucketsize / 2.0);
	}

	$buckets{$bucket}++;
    }

    foreach my $key (keys %buckets) {
	print DAT "$key $buckets{$key}\n";
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile_lg") || panic "Failed $epsfile_lg\n";

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Individual percentage error histogram}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EI_{avg%%}=%.2f%%, EI_{rms%%}=%.2f%%\"\n", $DeltaTRbar, $EIavgpct, $EIrmspct;
    print GNU "set xrange [$xrange_min:]\n";
    print GNU "set yrange [0:$N]\n";
    print GNU "set xlabel \"[{/Symbol D}T_C(n) {/Symbol - D}T_R(n)] {/Symbol \\270 D}T_R(n) {/Symbol \\264} 100.0, units=%\"\n";
    print GNU "set ylabel 'Number of requests (N=$N)'\n";
    print GNU "set boxwidth $bucketsize\n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3\n";
    print GNU "quit\n";

    close(GNU);

    ## small version

    open(GNU, "|gnuplot > $epsfile_sm") || panic "Failed $epsfile_sm\n";

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Individual percentage error histogram}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EI_{avg%%}=%.2f%%, EI_{rms%%}=%.2f%%\"\n", $DeltaTRbar, $EIavgpct, $EIrmspct;
    print GNU "set xrange [-25:25]\n";
    print GNU "set yrange [0:$N]\n";
    print GNU "set xlabel \"[{/Symbol D}T_C(n) {/Symbol - D}T_R(n)] {/Symbol \\270 D}T_R(n) {/Symbol \\264} 100.0, units=%\"\n";
    print GNU "set ylabel 'Number of requests (N=$N)'\n";
    print GNU "set boxwidth $bucketsize\n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3\n";
    print GNU "quit\n";

    close(GNU);
}



sub process_min {
    my ($requests, $which) = @_;
    my $min = @$requests[0]->{$which};
    foreach my $request (@$requests) {
	$min = $request->{$which} if ($min > $request->{$which});
    }
    return $min;
}


sub process_max {
    my ($requests, $which) = @_;
    my $max = @$requests[0]->{$which};
    foreach my $request (@$requests) {
	$max = $request->{$which} if ($max < $request->{$which});
    }
    return $max;
}


sub process_avg {
    my ($requests, $which) = @_;
    my $sum = 0;
    foreach my $request (@$requests) {
	$sum += $request->{$which};
    }
    return ($sum/@$requests);
}


sub process_rms {
    my ($requests, $which) = @_;
    my $sum_squares = 0;
    foreach my $request (@$requests) {
	$sum_squares += $request->{$which} * $request->{$which};
    }
    return sqrt($sum_squares/@$requests);
}


sub process_aggregate {
    my ($prefix, $requests) = @_;

    my @sorted = process_aggregate_sort(\@$requests);

    process_aggregate_stats_distribution($prefix, \@sorted, "hos", "sim");
    process_aggregate_stats_distribution($prefix, \@sorted, "mem", "sim");
    process_aggregate_stats_distribution($prefix, \@sorted, "hos", "mem");
}


sub process_aggregate_sort {
    my ($requests) = @_;

    my @sorted = ();

    my @hos = ();
    my @mem = ();
    my @sim = ();

    foreach $request (@$requests) {
	push @hos, $request->{hos_ms};
	push @mem, $request->{mem_ms};
	push @sim, $request->{sim_ms};
    }

    sub numerically { $a <=> $b };
    @hos = sort numerically @hos;
    @mem = sort numerically @mem;
    @sim = sort numerically @sim;

    for (my $i = 0; $i < @$requests; $i++) {
	my $hos_ms = $hos[$i];
	my $mem_ms = $mem[$i];
	my $sim_ms = $sim[$i];

	my $hos_sim_err_ms  = $hos_ms - $sim_ms;
	my $hos_sim_err_pct = $hos_sim_err_ms / $sim_ms * 100.0;

	my $mem_sim_err_ms  = $mem_ms - $sim_ms;
	my $mem_sim_err_pct = $mem_sim_err_ms / $sim_ms * 100.0;

	my $hos_mem_err_ms  = $hos_ms - $mem_ms;
	my $hos_mem_err_pct = $hos_mem_err_ms / $mem_ms * 100.0;

	push @sorted, {
	    hos_ms => $hos_ms,
	    mem_ms => $mem_ms,
	    sim_ms => $sim_ms,
	    hos_sim_err_ms  => $hos_sim_err_ms,
	    hos_sim_err_pct => $hos_sim_err_pct,
	    mem_sim_err_ms  => $mem_sim_err_ms,
	    mem_sim_err_pct => $mem_sim_err_pct,
	    hos_mem_err_ms  => $hos_mem_err_ms,
	    hos_mem_err_pct => $hos_mem_err_pct,
	};
    }

    return @sorted;
}
    

sub process_aggregate_stats_distribution {
    my ($prefix, $requests, $left, $right) = @_;

    my $N = @$requests;
    my $DeltaTRbar = process_avg(\@$requests, "$right\_ms");

    my $EArms = process_rms(\@$requests, "$left\_$right\_err_ms");
    my $EArmspct = $EArms / $DeltaTRbar * 100.0;

    my $abbrv_EArms = (int($EArms * 1000.0)) / 1000.0;
    my $abbrv_EArmspct = (int($EArmspct * 1000.0)) / 1000.0;

    ## Print the output.

    my $stats_name = "$prefix.agg.$left\_$right.stats";
    open(STATS, ">$stats_name" || panic "Cannot open $stats_name\n");

    printf STATS "N          = %d requests\n", $N;
    printf STATS "DeltaTRbar = %f ms\n", $DeltaTRbar;
    printf STATS "EArms      = %f ms\n", $EArms;
    printf STATS "EArmspct   = %f %%\n", $EArmspct;

    close (STATS);

#    system("emacs $stats_name &");

    ## Next, create the distribution

    $datfile = "$prefix.agg.$left\_$right.dat";
    $epsfile = "$prefix.agg.$left\_$right.eps";

    open(DAT, ">$datfile") || panic "$datfile";

    for (my $i = 0; $i < @$requests; $i++) {
	print DAT
	    (($i+1) / @$requests) . 
	    " @$requests->[$i]{\"$left\_ms\"}" .
	    " @$requests->[$i]{\"$right\_ms\"}\n"
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile") || panic "Failed gnuplot $epsfile\n";

    print GNU "set term postscript enhanced color \"Helvetica\" 24\n";
    printf GNU "set title \"{/Helvetica-Bold Aggregate request error distributions}\\n\\n\@^{\\305\\305\\305\\305\\305\\305\\305}{/Symbol D}T_R=%.3f ms, EA_{rms}=%.3f ms, EA_{rms%}=%.2f%%\"\n", $DeltaTRbar, $EArms, $EArmspct;
    print GNU "set yrange [0:1]\n";
    print GNU "set xlabel \"Time, units=ms\n";
    print GNU "set ylabel 'Fraction of requests (N=$N)'\n";
    print GNU "set key bottom\n";
    print GNU "plot '$datfile' using 2:1 title 'D_C ($left)' with lines lt 1 lw 3, '$datfile' using 3:1 title 'D_R ($right)' with lines lt 2 lw 3\n";
    print GNU "quit\n";

    close(GNU);

#    system("gv $epsfile &");
}


########################################################################


sub merge {
    my ($timestamp) = @_;
    my $prefix = "$results_dir/$timestamp/$timestamp";

    my @mem = merge_load_mem($prefix);
    my @sim = merge_load_sim($prefix);
    my @hos = merge_load_hos($prefix);

    merge_print($prefix, \@mem, \@sim, \@hos);
}


sub merge_load_mem {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE   = 31;
    my $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE = 32;

    open(MEM, "$prefix.mem") || panic "Cannot open $prefix.mem";

    my %arrivals = ();
    my @requests = ();

    while (<MEM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($opcode ne "08" 
	    and $opcode ne "0a"
	    and $opcode ne "28"
	    and $opcode ne "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE) {
	    $arrivals{$serial} = {
		offset    => $offset,
		length    => $length,
		opcode    => $opcode,
		timestamp => $timestamp,
	    };
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE) {
	    panic "mem offset" if ($arrivals{$serial}{offset} != $offset);
	    panic "mem length" if ($arrivals{$serial}{length} != $length);
	    panic "mem opcode" if ($arrivals{$serial}{opcode} != $opcode);
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		arrival    => $arrivals{$serial}{timestamp},
		completion => $timestamp,
	    };
	}
	else {
	    panic "mem Unknown code: $_";
	}
    }

    close(MEM);

    return @requests;
}


sub merge_load_sim {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    = 40;

    open(SIM, "$prefix.sim") || panic "Cannot open $prefix.sim";

    my @requests = ();

    while (<SIM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($code == $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME) {
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		time       => $timestamp,
	    };
	}
	else {
	    panic "sim Unknown code: $_";
	}
    }

    close(SIM);

    return @requests;
}


sub merge_load_hos {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE       = 21;
    my $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE     = 22;

    my %arrivals = ();
    my @requests = ();

    open(HOS, "$prefix.hos") || panic "Cannot open $prefix.hos";

    while (<HOS>) {
	my ($offset, $length, $opcode, $code, $timestamp) = split;

	if ($opcode ne "08" 
	    and $opcode ne "0a"
	    and $opcode ne "28"
	    and $opcode ne "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE) {
	    push @{$arrivals{($offset, $length, $opcode)}}, $timestamp;
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE) {
	    my $arrival = shift @{$arrivals{($offset, $length, $opcode)}}; 
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		arrival    => $arrival,
		completion => $timestamp,
	    };
	}
	else {
	    panic "hos Unknown code: $_";
	}
    }

    close(HOS);

    return @requests;
}


sub merge_hos_only {
    my ($timestamp) = @_;

    my $prefix = "$results_dir/$timestamp/$timestamp";

    my @hos = merge_load_hos($prefix);

    open(MERGE, ">$prefix.merge") || die "Cannot write $prefix.merge\n";

    for (my $i = 0; @hos != 0; $i++) {
	my $hos_elem = shift(@hos);
	my $hos_ms = ($hos_elem->{completion} - $hos_elem->{arrival}) / $host_cycles * 1000.0;

	printf MERGE "%6d %8d %4d %2d ",
	    $i + 1,
	    $hos_elem->{offset},
	    $hos_elem->{length},
	    $hos_elem->{opcode};
	print MERGE $hos_elem->{arrival} . " " . $hos_elem->{completion};
	printf MERGE "%7.3f\n",
            $hos_ms;
    }

    close(MERGE);
}


sub merge_print {
    my ($prefix, $mem_rqsts, $sim_rqsts, $hos_rqsts) = @_;

    my @mem_requests = @$mem_rqsts;
    my @sim_requests = @$sim_rqsts;
    my @hos_requests = @$hos_rqsts;

    my $v1 = @hos_requests;
    my $v2 = @mem_requests;
    my $v3 = @sim_requests;
    print "hos $v1, mem $v2, sim $v3\n";

    panic "FATAL: Mismatched sizes" if (@mem_requests != @sim_requests
					or @mem_requests != @hos_requests);

    open(MERGE, ">$prefix.merge") || panic "Cannot write $prefix.merge\n";
#    autoflush MERGE 1;

    for (my $i = 0; @mem_requests != 0; $i++) {

	my $mem_elem = shift(@mem_requests);
	my $hos_elem = shift(@hos_requests);

	panic "mem-hos mismatch at $i\n" if 
	    ($mem_elem->{offset} != $hos_elem->{offset}
	     or $mem_elem->{length} != $hos_elem->{length}
	     or $mem_elem->{opcode} != $hos_elem->{opcode});

	my @sim_stack = ();
	my $sim_elem = shift(@sim_requests);
	while ($mem_elem->{offset} != $sim_elem->{offset}
	       or $mem_elem->{length} != $sim_elem->{length}) {
	    die "Cannot locate sim request\n" if (@sim_requests == 0);
	    push @sim_stack, $sim_elem;
	    $sim_elem = shift(@sim_requests);
	}
	@sim_requests = (@sim_stack, @sim_requests);

#	if ($mem_elem->{opcode} eq "08"        ## READ 6
#	    or $mem_elem->{opcode} eq "28") {  ## READ 10
#	if ($mem_elem->{opcode} eq "0a"        ## WRITE 6
#	    or $mem_elem->{opcode} eq "2a") {  ## WRITE 10
	{
	  print MERGE
	    $i + 1                           . " " . 
	    $mem_elem->{offset}     . " " . 
	    $mem_elem->{length}     . " " .
	    $mem_elem->{opcode}     . " " .
	    $hos_elem->{arrival}    . " " . 
	    $hos_elem->{completion} . " " .
	    $mem_elem->{arrival}    . " " . 
	    $mem_elem->{completion} . " " .
	    $sim_elem->{time}       . "\n";
        }
    }

    my $v1 = @hos_requests;
    my $v2 = @mem_requests;
    my $v3 = @sim_requests;
    print "hos $v1, mem $v2, sim $v3\n";

    die "Nonzero mem array!\n" if (@mem_requests != 0);
    die "Nonzero sim array!\n" if (@sim_requests != 0);
    die "Nonzero hos array!\n" if (@hos_requests != 0);

    close (MERGE);
}

########################################################################


sub make_statistics_hash {
    # Called with the list of timestamps as its parameter

    my %rethash = ();

    foreach $timestamp (@_) {
	foreach $which ("mem_sim", "hos_mem", "hos_sim") {
	    my $statsfile = "$results_dir/$timestamp/$timestamp.per.$which.stats";
	    open (STATS, $statsfile) || die "$statsfile";
	    while (<STATS>) {
		my ($variable, $equals, $value, $units) = split;
		$rethash{$timestamp}{$which}{$variable} = $value;
	    }
	    close (STATS);

	    my $statsfile = "$results_dir/$timestamp/$timestamp.agg.$which.stats";
	    open (STATS, $statsfile) || die "$statsfile";
	    while (<STATS>) {
		my ($variable, $equals, $value, $units) = split;
		if ($variable eq "N" or $variable eq "DeltaTRbar") {
		    die if ($rethash{$timestamp}{$which}{$variable} != $value);
		} else {
		    $rethash{$timestamp}{$which}{$variable} = $value;
		}
	    }
	    close (STATS);
	}
    }

    return %rethash;
}


sub make_tables {
    my ($runid) = @_;

    my @source = ();
    my $sourcefile = "data/$runid/source";

    open (SOURCE, "$sourcefile") || die "$sourcefile";

    my @stamps = ();
    my %typemap = ();

    while (<SOURCE>) {
	if (!(/^%/)) {  # if line doesn't begin with %
	    /.+?\{(.+?)\}\{(\d+)\} % (.+?) (\w+)/;
	    # $1 is the latex name for the run (\timingEigWA)
	    # $2 is the timestamp
	    # $3 is the trace type (t-8-r)
	    # $4 is the adapter (adaptec)
	    push @stamps, $2;
	    $typemap{$3}{$4} = $2;
	}
    }

    close (SOURCE);

    my %values = make_statistics_hash(@stamps);

    foreach $which ("mem_sim", "hos_mem", "hos_sim") {
	my $typestr = $which;
	$typestr =~ s/_/\$\\rightarrow\$/;

	my $tablefile = "data/$runid/table.$which";
	open (TABLE, ">$tablefile") || die "$tablefile";
	print TABLE "\\begin{tabular}{|l|c" . "||c|c"x3 . "|} \\hline\n";
	print TABLE "\\DeltaTC\$\\rightarrow\$\\DeltaTR & \\DeltaTRbar & \\eiavg & \\eiavgpct & \\eirms & \\eirmspct & \\earms & \\earmspct \\\\\n";
	print TABLE "($typestr) & (ms) & (ms) & (\\%) & (ms) & (\\%) & (ms) & (\\%) \\\\ \\hline\n";

	foreach $hba ("adaptec", "qlogic") {
	    printf TABLE "\\hline \\multicolumn{8}{|l|}{%s} \\\\ \\hline\n", ucfirst($hba);

	    foreach $size (2, 8, 64, 128) {
		foreach $rw ("read", "write") {
		    print TABLE "\\hline\n";
		    my $type    = "t-$size-" . ($rw eq "read" ? "r" : "w");
		    my $timestamp = $typemap{$type}{$hba};

		    $output = sprintf "  %s & %.3f & %.3f & %.2f & %.3f & %.2f & %.3f & %.2f \\\\\n",
			"$size block $rw",
			$values{$timestamp}{$which}{DeltaTRbar},
			$values{$timestamp}{$which}{EIavg},
			$values{$timestamp}{$which}{EIavgpct},
			$values{$timestamp}{$which}{EIrms},
			$values{$timestamp}{$which}{EIrmspct},
			$values{$timestamp}{$which}{EArms},
			$values{$timestamp}{$which}{EArmspct};

		    $output =~ s/-/--/g;
		    print TABLE $output;
		}
		print TABLE "\\hline\n";
	    }
	}
	print TABLE "\\end{tabular}\n";
	close (TABLE);
    }
}


sub make_summary {
    my ($runid) = @_;

    chdir "data/$runid" or die "Can't cd to data/$runid: $!\n";
    system "cp ../../summary.tex .";
    system "latex -interaction=batchmode summary.tex";
    system "latex -interaction=batchmode summary.tex";
    system "dvips -q -o summary.ps summary.dvi";
    system "rm summary.aux summary.log";
    chdir "../.." or die "Can't cd to original directory: $!\n";
}

########################################################################

1;
