#!/usr/bin/perl

merge("../results/1086363307");

sub merge {
    my ($prefix) = @_;

    my @mem = merge_load_mem($prefix);
    my @sim = merge_load_sim($prefix);
    my @hos = merge_load_hos($prefix);

    merge_print($prefix, \@mem, \@sim, \@hos);
}


sub merge_load_mem {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE   = 31;
    my $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE = 32;

    open(MEM, "$prefix.mem") || die "Cannot open mem file";

    my %arrivals = ();
    my @requests = ();

    while (<MEM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($opcode != "08" 
	    and $opcode != "0a"
	    and $opcode != "28"
	    and $opcode != "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE) {
	    $arrivals{$serial} = {
		offset    => $offset,
		length    => $length,
		opcode    => $opcode,
		timestamp => $timestamp,
	    };
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE) {
	    die "FATAL: mem offset" if ($arrivals{$serial}{offset} != $offset);
	    die "FATAL: mem length" if ($arrivals{$serial}{length} != $length);
	    die "FATAL: mem opcode" if ($arrivals{$serial}{opcode} != $opcode);
	    push @requests, {
		offset     => $offset,
		length     => $length,
		code       => $code,
		serial     => $serial,
		arrival    => $arrivals{$serial}{timestamp},
		completion => $timestamp,
	    };
	}
	else {
	    die "FATAL: mem Unknown code: $_";
	}
    }

    close(MEM);

    return @requests;
}


sub merge_load_sim {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    = 40;

    open(SIM, "$prefix.sim") || die "Cannot open sim file";

    my @requests = ();

    while (<SIM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($code == $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME) {
	    push @requests, {
		offset     => $offset,
		length     => $length,
		code       => $code,
		serial     => $serial,
		time       => $timestamp,
	    };
	}
	else {
	    die "FATAL: sim Unknown code: $_";
	}
    }

    close(SIM);

    return @requests;
}


sub merge_load_hos {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE       = 21;
    my $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE     = 22;

    my %arrivals = ();
    my @requests = ();

    open(HOS, "$prefix.hos") || die "Cannot open hos file";

    while (<HOS>) {
	my ($offset, $length, $opcode, $code, $timestamp) = split;

	if ($opcode != "08" 
	    and $opcode != "0a"
	    and $opcode != "28"
	    and $opcode != "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE) {
	    push @{$arrivals{($offset, $length, $opcode)}}, $timestamp;
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE) {
	    my $arrival = shift @{$arrivals{($offset, $length, $opcode)}}; 
	    push @requests, {
		offset     => $offset,
		length     => $length,
		code       => $code,
		serial     => $serial,
		arrival    => $arrival,
		completion => $timestamp,
	    };
	}
	else {
	    die "FATAL: hos Unknown code: $_";
	}
    }

    close(HOS);

    return @requests;
}


sub merge_print {
    my ($prefix, $mem_requests, $sim_requests, $hos_requests) = @_;

    die "FATAL: Mismatched sizes" if (@$mem_requests != @$sim_requests
			       or @$mem_requests != @$hos_requests);

    open(MERGE, ">$prefix.merge") || die "Cannot write $prefix.merge\n";

    for (my $i = 0; $i < @$mem_requests; $i++) {
	die "FATAL: Mismatched requests" if
	    (@$mem_requests[$i]->{offset} != @$sim_requests[$i]->{offset}
	     or @$mem_requests[$i]->{length} != @$sim_requests[$i]->{length}
	     or @$mem_requests[$i]->{offset} != @$hos_requests[$i]->{offset}
	     or @$mem_requests[$i]->{length} != @$hos_requests[$i]->{length}
	     or @$mem_requests[$i]->{opcode} != @$hos_requests[$i]->{opcode});

	print MERGE
	    $i + 1                           . " " . 
	    @$mem_requests[$i]->{offset}     . " " . 
	    @$mem_requests[$i]->{length}     . " " .
	    @$hos_requests[$i]->{arrival}    . " " . 
	    @$hos_requests[$i]->{completion} . " " .
	    @$mem_requests[$i]->{arrival}    . " " . 
	    @$mem_requests[$i]->{completion} . " " .
	    @$sim_requests[$i]->{time}       . "\n";
    }

    close (MERGE);
}

