
########################################################################

sub log_entry {
    my ($msg) = @_;
    #print LOG "$msg";
    print "$msg";
}

sub panic {
    my ($msg) = @_;
    print "###############################################################\n";
    print "########################### FAILURE ###########################\n";
    print "###############################################################\n";
    log_entry($msg);
    die "$msg";
}


sub ssh_cmd {
    my ($destination,$command) = @_;

    log_entry("ssh $destination $command\n");

    # fork a process for the ssh
    my $pid = open(KID, "-|");
    panic "unable to fork" unless defined $pid;

    if ($pid == 0) {  # child process
	open STDERR,'>&STDOUT';
	@cmd = ("ssh","$destination","$command");
	exec(@cmd);
	# NOTREACHED
    } else {          # parent process
	while ($output = <KID>) {
	    log_entry("ssh: $output");
	}
	close KID;
	return $?;
    }
}

sub ssh_cmd_checkresult {
    my ($destination,$command,$expected_result) = @_;

    my $result = ssh_cmd($destination,$command);

    if ($result != $expected_result) {
	panic("bad ssh result ($result != $expected_result)");
    }
}

########################################################################

sub targ_start_memulator_times {
    my ($timestamp) = @_;
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -b $targ_memulator_store -c -f $results_dir/$timestamp $targ_memulator_address &) < /dev/null >& /dev/null");
    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -c -f $results_dir/$timestamp -z 35566478 $targ_memulator_address &) < /dev/null >& $results_dir/$timestamp.mem.log");
    sleep 1;
}


sub targ_start_memulator_notimes {
#    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -b $targ_memulator_store $targ_memulator_address &) < /dev/null >& /dev/null");
    ssh_cmd("$targ_name", "(cd /home/griffin2/proj/memulator/src/memulator ; ./memulator -z 35566478 $targ_memulator_address &) < /dev/null >& /dev/null");
    sleep 1;
}


sub targ_verify_memulator {
    ssh_cmd_checkresult("$targ_name", "ps ax | grep memulator | grep -v grep", 0);
}


sub host_insmod_qla2200 {
    ssh_cmd_checkresult("root\@$host_name", "insmod qla2200", 0);
}


sub host_add_single_device {
    ssh_cmd_checkresult("root\@$host_name", "echo 'scsi add-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi", 0);
}


sub host_insmod_hosttimes {
    my ($timestamp) = @_;
    ssh_cmd_checkresult("root\@$host_name", "insmod /home/griffin2/workbench/hosttimes/hosttimes.o filename=$results_dir/$timestamp.hos host_no=$host_address_host_no channel=$host_address_channel target=$host_address_target lun=$host_address_lun", 0);
}

########################################################################

sub host_rmmod_hosttimes {
    ssh_cmd("root\@$host_name", "rmmod hosttimes");
}


sub host_rmmod_qla2200 {
    ssh_cmd("root\@$host_name", "rmmod qla2200");
}


sub host_remove_single_device {
#    ssh_cmd_checkresult("root\@$host_name", "echo 'scsi remove-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi", 256);
    ssh_cmd("root\@$host_name", "echo 'scsi remove-single-device $host_address_host_no $host_address_channel $host_address_target $host_address_lun' >> /proc/scsi/scsi");
}


sub targ_stop_memulator {
    ssh_cmd_checkresult("root\@$targ_name", "killall -1 memulator", 0);
}

########################################################################

sub config_qlogic {
    $host_name = "nc93.pdl.cmu.edu";  # host machine
    $targ_name = "nc94.pdl.cmu.edu";  # target (emulation) machine

    $host_cycles = 1993907171.0;
    $targ_cycles = 1993868287.0;

    $targ_memulator_store = "/dev/da1";  # backing store for emulated disk
    $targ_memulator_address = "1:94:0";  # bus:target:lun for emualted disk

    $host_address_host_no = 3;
    $host_address_channel = 0;
    $host_address_target  = 0;
    $host_address_lun     = 0;

    $main_dir = "/home/griffin2/workbench";
    $results_dir = "$main_dir/results";
}


sub config_adaptec {
    $host_name = "nc93.pdl.cmu.edu";  # host machine
    $targ_name = "nc94.pdl.cmu.edu";  # target (emulation) machine

    $host_cycles = 1993907171.0;
    $targ_cycles = 1993868287.0;

    $targ_memulator_store = "/dev/da1";  # backing store for emulated disk
    $targ_memulator_address = "2:4:0";   # bus:target:lun for emualted disk

    $host_address_host_no = 1;
    $host_address_channel = 0;
    $host_address_target  = 4;
    $host_address_lun     = 0;

    $main_dir = "/home/griffin2/workbench";
    $results_dir = "$main_dir/results";
}


sub qlogic_start {
    my ($timestamp) = @_;

    config_qlogic();

    host_rmmod_qla2200();
    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_insmod_qla2200();
    targ_stop_memulator();

    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub qlogic_stop {
    config_qlogic();

    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_rmmod_qla2200();
    targ_stop_memulator();
}


sub adaptec_start {
    my ($timestamp) = @_;

    config_adaptec();

    targ_start_memulator_notimes();
    targ_verify_memulator();
    host_add_single_device();
    targ_stop_memulator();

    targ_start_memulator_times($timestamp);
    targ_verify_memulator();
    host_insmod_hosttimes($timestamp);
}


sub adaptec_stop {
    config_adaptec();

    host_rmmod_hosttimes();
    targ_stop_memulator();

    targ_start_memulator_notimes();
    host_remove_single_device();
    targ_stop_memulator();
}

########################################################################

sub process_data {
    my ($timestamp) = @_;
    my $prefix = "$results_dir/$timestamp";

    my @requests = process_merge_file($prefix);
    process_print_datfile($prefix, \@requests);

    process_per_request($prefix, \@requests);
    process_aggregate($prefix, \@requests);
}


sub process_merge_file {
    my ($prefix) = @_;

    my $fh_name = "$prefix.merge";
    open(FH, "$fh_name") || panic "Cannot open $fh_name";

    my @requests = ();

    while (<FH>) {
	my ($num, 
	    $offset, $length, 
	    $opcode, 
	    $hos_arrive, $hos_complete, 
	    $mem_arrive, $mem_complete, 
	    $simtime) = split;

	my $hos_ms = ($hos_complete - $hos_arrive) / $host_cycles * 1000.0;
	my $mem_ms = ($mem_complete - $mem_arrive) / $targ_cycles * 1000.0;
	my $sim_ms = $simtime;

	my $hos_sim_err_ms  = $hos_ms - $sim_ms;
	my $hos_sim_err_pct = $hos_sim_err_ms / $sim_ms * 100.0;

	my $mem_sim_err_ms  = $mem_ms - $sim_ms;
	my $mem_sim_err_pct = $mem_sim_err_ms / $sim_ms * 100.0;

	my $hos_mem_err_ms  = $hos_ms - $mem_ms;
	my $hos_mem_err_pct = $hos_mem_err_ms / $mem_ms * 100.0;

	push @requests, {
	    hos_ms => $hos_ms,
	    mem_ms => $mem_ms,
	    sim_ms => $sim_ms,
	    hos_sim_err_ms  => $hos_sim_err_ms,
	    hos_sim_err_pct => $hos_sim_err_pct,
	    mem_sim_err_ms  => $mem_sim_err_ms,
	    mem_sim_err_pct => $mem_sim_err_pct,
	    hos_mem_err_ms  => $hos_mem_err_ms,
	    hos_mem_err_pct => $hos_mem_err_pct,
	};
    }

    close(FH);

    return @requests;
}


sub process_print_datfile {
    my ($prefix, $requests) = @_;

    my $out_name = "$prefix.dat";
    open(OUT, ">$out_name") || panic "Cannot open $out_name";

    for (my $i = 0; $i < @$requests; $i++) {
	printf OUT 
	    "%d %7.3f %7.3f %7.3f %7.3f %7.3f%% %7.3f %7.3f%% %7.3f %7.3f%%\n",
	    $i + 1,
	    @$requests[$i]->{hos_ms},
	    @$requests[$i]->{mem_ms},
	    @$requests[$i]->{sim_ms},
	    @$requests[$i]->{hos_sim_err_ms },
	    @$requests[$i]->{hos_sim_err_pct},
	    @$requests[$i]->{mem_sim_err_ms },
	    @$requests[$i]->{mem_sim_err_pct},
	    @$requests[$i]->{hos_mem_err_ms },
	    @$requests[$i]->{hos_mem_err_pct};
    }

    close(OUT);
}


sub process_per_request {
    my ($prefix, $requests) = @_;

    process_per_request_stats($prefix, \@$requests, "hos", "sim");
    process_per_request_stats($prefix, \@$requests, "mem", "sim");
    process_per_request_stats($prefix, \@$requests, "hos", "mem");

    process_histogram_error($prefix, \@$requests, "hos_sim");
    process_histogram_error($prefix, \@$requests, "mem_sim");
    process_histogram_error($prefix, \@$requests, "hos_mem");

    process_histogram_percent_error($prefix, \@$requests, "hos_sim");
    process_histogram_percent_error($prefix, \@$requests, "mem_sim");
    process_histogram_percent_error($prefix, \@$requests, "hos_mem");
}


sub process_per_request_stats {
    my ($prefix, $requests, $comparitor, $reference) = @_;

    my $N = @$requests;
    my $DeltaTRbar = process_avg(\@$requests, "$reference\_ms");

    my $EIavg = process_avg(\@$requests, "$comparitor\_$reference\_err_ms");
    my $EIavgpct = $EIavg / $DeltaTRbar * 100.0;

    my $EIrms = process_rms(\@$requests, "$comparitor\_$reference\_err_ms");
    my $EIrmspct = $EIrms / $DeltaTRbar * 100.0;

    my $fh_name = "$prefix.per.$comparitor\_$reference.stats";
    open(FH, ">$fh_name" || panic "Cannot open $fh_name\n");

    printf FH "N          = %d requests\n", $N;
    printf FH "DeltaTRbar = %f ms\n", $DeltaTRbar;
    printf FH "EIavg      = %f ms\n", $EIavg;
    printf FH "EIavgpct   = %f %%\n", $EIavgpct;
    printf FH "EIrms      = %f ms\n", $EIrms;
    printf FH "EIrmspct   = %f %%\n", $EIrmspct;

    close(FH);

    system("emacs $fh_name &");
}


sub process_min {
    my ($requests, $which) = @_;
    my $min = @$requests[0]->{$which};
    foreach my $request (@$requests) {
	$min = $request->{$which} if ($min > $request->{$which});
    }
    return $min;
}


sub process_max {
    my ($requests, $which) = @_;
    my $max = @$requests[0]->{$which};
    foreach my $request (@$requests) {
	$max = $request->{$which} if ($max < $request->{$which});
    }
    return $max;
}


sub process_avg {
    my ($requests, $which) = @_;
    my $sum = 0;
    foreach my $request (@$requests) {
	$sum += $request->{$which};
    }
    return ($sum/@$requests);
}


sub process_rms {
    my ($requests, $which) = @_;
    my $sum_squares = 0;
    foreach my $request (@$requests) {
	$sum_squares += $request->{$which} * $request->{$which};
    }
    return sqrt($sum_squares/@$requests);
}


sub process_histogram_error {
    my ($prefix, $requests, $which) = @_;

    my $datfile = "$prefix.per.$which.err.dat";
    my $epsfile = "$prefix.per.$which.err.eps";
    my $bucketsize = 0.01;
    my %buckets = ();

    open(DAT, "|sort -g >$datfile") || panic "Cannot open $datfile\n";

    foreach my $request (@$requests) {
	my $bucket = int($request->{"$which\_err_ms"} / $bucketsize);
	$buckets{$bucket}++;
    }

    foreach my $key (keys %buckets) {
	print DAT ($key * $bucketsize) . " $buckets{$key}\n";
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile") || panic "Failed gnuplot $epsfile\n";

    print GNU "set term postscript landscape color \"Helvetica\" 18 \n";
    print GNU "set title 'Distribution of service time error (" . @$requests . " requests)' \n";
    print GNU "set yrange [0:] \n";
    print GNU "set xlabel 'Measured time - simulated time (milliseconds)' \n";
    print GNU "set ylabel 'Number of requests' \n";
    print GNU "set boxwidth $bucketsize \n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3 \n";
    print GNU "quit\n";

    close(GNU);

    system("gv $epsfile &");
}


sub process_histogram_percent_error {
    my ($prefix, $requests, $which) = @_;

    my $datfile = "$prefix.per.$which.pct.dat";
    my $epsfile = "$prefix.per.$which.pct.eps";
    my $bucketsize = 0.01;
    my %buckets = ();

    open(DAT, "|sort -g >$datfile") || panic "Cannot open $datfile\n";

    foreach my $request (@$requests) {
	my $bucket = int($request->{"$which\_err_pct"} / $bucketsize);
	$buckets{$bucket}++;
    }

    foreach my $key (keys %buckets) {
	print DAT ($key * $bucketsize) . " $buckets{$key}\n";
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile") || panic "Failed gnuplot $epsfile\n";

    print GNU "set term postscript landscape color \"Helvetica\" 18 \n";
    print GNU "set title 'Distribution of service time percent error' \n";
    print GNU "set yrange [0:] \n";
    print GNU "set xlabel 'Percent error (%)' \n";
    print GNU "set ylabel 'Number of requests' \n";
    print GNU "set boxwidth $bucketsize \n";
    print GNU "plot '$datfile' title '' with boxes lt -1 lw 3 \n";
    print GNU "quit\n";

    close(GNU);
}


sub process_aggregate {
    my ($prefix, $requests) = @_;

    my @sorted = process_aggregate_sort(\@$requests);

    process_aggregate_stats_distribution($prefix, \@sorted, "hos", "sim");
    process_aggregate_stats_distribution($prefix, \@sorted, "mem", "sim");
    process_aggregate_stats_distribution($prefix, \@sorted, "hos", "mem");
}


sub process_aggregate_sort {
    my ($requests) = @_;

    my @sorted = ();

    my @hos = ();
    my @mem = ();
    my @sim = ();

    foreach $request (@$requests) {
	push @hos, $request->{hos_ms};
	push @mem, $request->{mem_ms};
	push @sim, $request->{sim_ms};
    }

    sub numerically { $a <=> $b };
    @hos = sort numerically @hos;
    @mem = sort numerically @mem;
    @sim = sort numerically @sim;

    for (my $i = 0; $i < @$requests; $i++) {
	my $hos_ms = $hos[$i];
	my $mem_ms = $mem[$i];
	my $sim_ms = $sim[$i];

	my $hos_sim_err_ms  = $hos_ms - $sim_ms;
	my $hos_sim_err_pct = $hos_sim_err_ms / $sim_ms * 100.0;

	my $mem_sim_err_ms  = $mem_ms - $sim_ms;
	my $mem_sim_err_pct = $mem_sim_err_ms / $sim_ms * 100.0;

	my $hos_mem_err_ms  = $hos_ms - $mem_ms;
	my $hos_mem_err_pct = $hos_mem_err_ms / $mem_ms * 100.0;

	push @sorted, {
	    hos_ms => $hos_ms,
	    mem_ms => $mem_ms,
	    sim_ms => $sim_ms,
	    hos_sim_err_ms  => $hos_sim_err_ms,
	    hos_sim_err_pct => $hos_sim_err_pct,
	    mem_sim_err_ms  => $mem_sim_err_ms,
	    mem_sim_err_pct => $mem_sim_err_pct,
	    hos_mem_err_ms  => $hos_mem_err_ms,
	    hos_mem_err_pct => $hos_mem_err_pct,
	};
    }

    return @sorted;
}
    

sub process_aggregate_stats_distribution {
    my ($prefix, $requests, $left, $right) = @_;

    my $N = @$requests;
    my $DeltaTRbar = process_avg(\@$requests, "$right\_ms");

    my $EArms = process_rms(\@$requests, "$left\_$right\_err_ms");
    my $EArmspct = $EArms / $DeltaTRbar * 100.0;

    my $abbrv_EArms = (int($EArms * 1000.0)) / 1000.0;
    my $abbrv_EArmspct = (int($EArmspct * 1000.0)) / 1000.0;

    ## Print the output.

    my $stats_name = "$prefix.agg.$left\_$right.stats";
    open(STATS, ">$stats_name" || panic "Cannot open $stats_name\n");

    printf STATS "N          = %d requests\n", $N;
    printf STATS "DeltaTRbar = %f ms\n", $DeltaTRbar;
    printf STATS "EArms      = %f ms\n", $EArms;
    printf STATS "EArmspct   = %f %%\n", $EArmspct;

    close (STATS);

    system("emacs $stats_name &");

    ## Next, create the distribution

    $datfile = "$prefix.agg.$left\_$right.dat";
    $epsfile = "$prefix.agg.$left\_$right.eps";

    open(DAT, ">$datfile") || panic "$datfile";

    for (my $i = 0; $i < @$requests; $i++) {
	print DAT
	    (($i+1) / @$requests) . 
	    " @$requests->[$i]{\"$left\_ms\"}" .
	    " @$requests->[$i]{\"$right\_ms\"}\n"
    }

    close(DAT);

    open(GNU, "|gnuplot > $epsfile") || panic "Failed gnuplot $epsfile\n";

    print GNU "set term postscript landscape color \"Helvetica\" 18 \n";
    print GNU "set title 'Response time distributions, " . @$requests . " requests (EArms $abbrv_EArms ms [$abbrv_EArmspct%])' \n";
    print GNU "set xrange [0:30] \n";
    print GNU "set yrange [0:1] \n";
    print GNU "set xlabel 'Response time (milliseconds)' \n";
    print GNU "set ylabel 'Fraction of requests' \n";
    print GNU "set key bottom \n";
    print GNU "plot '$datfile' using 2:1 title '$left' with lines, '$datfile' using 3:1 title '$right' with lines \n";
    print GNU "quit\n";

    close(GNU);

    system("gv $epsfile &");
}



########################################################################


sub merge {
    my ($timestamp) = @_;
    my $prefix = "$results_dir/$timestamp";

    my @mem = merge_load_mem($prefix);
    my @sim = merge_load_sim($prefix);
    my @hos = merge_load_hos($prefix);

    merge_print($prefix, \@mem, \@sim, \@hos);
}


sub merge_load_mem {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE   = 31;
    my $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE = 32;

    open(MEM, "$prefix.mem") || panic "Cannot open $prefix.mem";

    my %arrivals = ();
    my @requests = ();

    while (<MEM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($opcode != "08" 
	    and $opcode != "0a"
	    and $opcode != "28"
	    and $opcode != "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE) {
	    $arrivals{$serial} = {
		offset    => $offset,
		length    => $length,
		opcode    => $opcode,
		timestamp => $timestamp,
	    };
	}
	elsif ($code == $MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE) {
	    panic "mem offset" if ($arrivals{$serial}{offset} != $offset);
	    panic "mem length" if ($arrivals{$serial}{length} != $length);
	    panic "mem opcode" if ($arrivals{$serial}{opcode} != $opcode);
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		arrival    => $arrivals{$serial}{timestamp},
		completion => $timestamp,
	    };
	}
	else {
	    panic "mem Unknown code: $_";
	}
    }

    close(MEM);

    return @requests;
}


sub merge_load_sim {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    = 40;

    open(SIM, "$prefix.sim") || panic "Cannot open $prefix.sim";

    my @requests = ();

    while (<SIM>) {
	my ($offset, $length, $opcode, $code, $timestamp, $serial) = split;

	if ($code == $MEMULATOR_EVENT_TARGET_SIMULATOR_TIME) {
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		time       => $timestamp,
	    };
	}
	else {
	    panic "sim Unknown code: $_";
	}
    }

    close(SIM);

    return @requests;
}


sub merge_load_hos {
    my ($prefix) = @_;

    my $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE       = 21;
    my $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE     = 22;

    my %arrivals = ();
    my @requests = ();

    open(HOS, "$prefix.hos") || panic "Cannot open $prefix.hos";

    while (<HOS>) {
	my ($offset, $length, $opcode, $code, $timestamp) = split;

	if ($opcode != "08" 
	    and $opcode != "0a"
	    and $opcode != "28"
	    and $opcode != "2a") {
	    # do nothing
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_ARRIVE) {
	    push @{$arrivals{($offset, $length, $opcode)}}, $timestamp;
	}
	elsif ($code == $MEMULATOR_EVENT_HOST_KERNEL_COMPLETE) {
	    my $arrival = shift @{$arrivals{($offset, $length, $opcode)}}; 
	    push @requests, {
		offset     => $offset,
		length     => $length,
		opcode     => $opcode,
		serial     => $serial,
		arrival    => $arrival,
		completion => $timestamp,
	    };
	}
	else {
	    panic "hos Unknown code: $_";
	}
    }

    close(HOS);

    return @requests;
}


sub merge_print {
    my ($prefix, $mem_requests, $sim_requests, $hos_requests) = @_;

    panic "FATAL: Mismatched sizes" if (@$mem_requests != @$sim_requests
					or @$mem_requests != @$hos_requests);

    open(MERGE, ">$prefix.merge") || panic "Cannot write $prefix.merge\n";

    for (my $i = 0; $i < @$mem_requests; $i++) {
	panic "FATAL: Mismatched requests" if
	    (@$mem_requests[$i]->{offset} != @$sim_requests[$i]->{offset}
	     or @$mem_requests[$i]->{length} != @$sim_requests[$i]->{length}
	     or @$mem_requests[$i]->{offset} != @$hos_requests[$i]->{offset}
	     or @$mem_requests[$i]->{length} != @$hos_requests[$i]->{length}
	     or @$mem_requests[$i]->{opcode} != @$hos_requests[$i]->{opcode});

	print MERGE
	    $i + 1                           . " " . 
	    @$mem_requests[$i]->{offset}     . " " . 
	    @$mem_requests[$i]->{length}     . " " .
	    @$mem_requests[$i]->{opcode}     . " " .
	    @$hos_requests[$i]->{arrival}    . " " . 
	    @$hos_requests[$i]->{completion} . " " .
	    @$mem_requests[$i]->{arrival}    . " " . 
	    @$mem_requests[$i]->{completion} . " " .
	    @$sim_requests[$i]->{time}       . "\n";
    }

    close (MERGE);
}

########################################################################

1;
